/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

char list[40][2] = 
    {
        /* ll_list[16][2] */ 
        {'a', 'a'}, {'a', 'c'}, {'a', 'g'}, {'a', 't'},
        {'c', 'a'}, {'c', 'c'}, {'c', 'g'}, {'c', 't'},
        {'g', 'a'}, {'g', 'c'}, {'g', 'g'}, {'g', 't'},
        {'t', 'a'}, {'t', 'c'}, {'t', 'g'}, {'t', 't'},
 
        /* ld_list[4][2] */ 
        {'a', 'A'}, {'c', 'C'}, {'g', 'G'}, {'t', 'T'},

        /* dd_list[4][2] */ 
        {'A', 'A'}, {'C', 'C'}, {'G', 'G'}, {'T', 'T'},

        /* dl_list[16][2] */ 
        {'A', 'a'}, {'A', 'c'}, {'A', 'g'}, {'A', 't'},
        {'C', 'a'}, {'C', 'c'}, {'C', 'g'}, {'C', 't'},
        {'G', 'a'}, {'G', 'c'}, {'G', 'g'}, {'G', 't'},
        {'T', 'a'}, {'T', 'c'}, {'T', 'g'}, {'T', 't'}
    };

/******************************************************************************
* GET_TRANSITIONS returns the transition matrix, a states list and states
* maps.
*   [big_mat list maps] = GET_TRANSITIONS(freq, comb) computes and returns
*   the transition matrix (1600x1600), a list representing the first states
*   of the first allele in letterspace (so a 40-vector , since the a
*   complete states is made of two alleles => 40x40 = 1600-vector), and
*   some more lists of the states in the maps structure.
*
*   N.B.: a capital letter represents a gap state with the memory of that
*   letter. For example, an 'A' represents the gap '-' with the memory that
*   the last letter seen is an a. So 'A' means '-_a', etc.
*
*   Parameters:
*   freq - a struct holding the fields:
*       freq.a - the frequency of 'a' nucleotides (e.g. 0.25)
*       freq.c - the frequency of 'c' nucleotides (e.g. 0.25)
*       freq.g - the frequency of 'g' nucleotides (e.g. 0.25)
*       freq.t - the frequency of 't' nucleotides (e.g. 0.25)
*       freq.gapopen - the penalty for gap open
*       freq.gapextend - the penalty for gap extend
*   comb - a struct holding the fields:
*       comb.c - a 5x8 matrix representing the 8 states where each of the 5
*           colors (0-3 and gap[4]) are present
*       comb.l - a 5x8 matrix representing the 8 states where each of the 5
*           possible letters (0-3 = {a,c,g,t} and gap[4]) are present
*
*   Returns:
*   big_mat - the 1600x1600 transition probability matrix
*   list - the 40x2 list of the first allele letters (e.g. {aa, ac, ...})
*   maps - a struct holding states to letter or states to color maps.
*       maps.full_letter - 1600x4 matrix, the letters of each of the 1600
*           states for both alleles. For example, the second state will be
*           aa/ac, which will be represented in ns as [0 0 0 1]
*       maps.call_letter - 1600x2 matrix, only the call letter of each of
*           the states for both alleles. For example, the second state will
*           be aa/ac, with the call letters a/c, which will be represented
*           in ns as [0 1]
*       maps.call_color - 1600x2 matrix, reprenting the call colors of each
*           of the states for both alleles. For example, the second state
*           will be aa/ac, which is 0/1 in colorspace, which will be
*           represented as [0 1]
*
*   Author: Adrian Dalca
*   TODO: not fully tested the final big_mat result...
*   NOTE: there currently isn't any snp penalty implemented in the
*       resulting transition matrix...
*******************************************************************************/
void get_transitions(
    Frequency_t    freq, 
    Comb_t         comb,
    VA_MatDouble * big_mat,
    Maps_t       * maps)
{
    struct timespec ts;
    struct timespec te;
    struct timespec tdiff;
    int rc = 0;
    VA_UINT32 yl;
    VA_UINT32 xl;
    VA_UINT32 x1;
    VA_UINT32 x2;
    VA_UINT32 y1;
    VA_UINT32 y2;

    char * letters = "acgt";
    char * gaps = "ACGT";
    int sz = sizeof(list) / sizeof(list[0]);
    int i;
    VA_MatDouble trans_mat = va_allocMatDouble(sz, sz);

    VA_ASSERT(big_mat != NULL);
    VA_ASSERT(maps != NULL);

    /* transitions */
    rc = clock_gettime(CLOCK_REALTIME, &ts);
    VA_ASSERT(rc == 0);

    fprintf(stderr, "\tBuilding Transition Matrix...\n");

    /* build first (primary) transition matrix (40 to 40 states) */

    /* set-up transition matrices*/
    memset(trans_mat.data, 0, trans_mat.size_x * 
        trans_mat.size_y * sizeof(VA_DOUBLE));

    /* go through each element in the lists */
    for (yl = 0; yl < sz; yl++)
    {
        char * y = list[yl];

        for (xl = 0; xl < sz; xl++)
        {
            char * x = list[xl];

            /* so now we have two vectors of length 2: the first state
             * y(1:2) and the second state (the one we are transitiong TO)
             * x(1:2)
             */

            /* if inner letters match and they are nucleotides (not gaps) */
            if ((y[1] == x[0]) && (strchr(letters, y[1]) != NULL))
            {
                /* normal to opening a gap, closing a gap to opening a gap */
                if (strchr(gaps, x[1]) != NULL)
                {
                    trans_mat.data[yl * trans_mat.size_x + xl] = freq.gapopen;
                }        
                else /* normal to normal, closing a gap to normal, */
                { 
                    switch (x[1])
                    {
                        case 'a':
                            trans_mat.data[yl * sz + xl] = freq.a;
                            break;
                        case 'c':
                            trans_mat.data[yl * sz + xl] = freq.c;
                            break;
                        case 'g':
                            trans_mat.data[yl * sz + xl] = freq.g;
                            break;
                        case 't':
                            trans_mat.data[yl * sz + xl] = freq.t;
                            break;
                    } 
                }
            }    
            /* if inner letters match and they are gaps     */
            else if (y[1] == x[0] && (strchr(gaps, y[1]) != NULL))
            {
                /* open gap to gap, or gap to gap */
                if (strchr(gaps, x[1]) != NULL)
                {
                    trans_mat.data[yl * trans_mat.size_x + xl] = freq.gapextend;
                }    
                else
                {    
                    /* gap to close cap, open gap to close gap */
                    switch (x[1])
                    {
                        case 'a':
                            trans_mat.data[yl * sz + xl] = freq.a;
                            break;
                        case 'c':
                            trans_mat.data[yl * sz + xl] = freq.c;
                            break;
                        case 'g':
                            trans_mat.data[yl * sz + xl] = freq.g;
                            break;
                        case 't':
                            trans_mat.data[yl * sz + xl] = freq.t;
                            break;
                    } 
                }
            }
        }
    }
       
    /* secondary transition matrix (40x40 to 40x40 states) */
      
    /* prepare the final transition mat. */
    *big_mat = va_allocMatDouble(sz * sz, sz * sz);

    /* state transitioning FROM: */
    /* the list index (1-40) of the first allele */
    for (y1 = 0; y1 < sz; y1++)
    {
        /* the list index (1-40) of the second allele */ 
        for (y2 = 0; y2 < sz; y2++)
        {
            /* state transitioning TO: */
            /* the list index (1-40) of the first allele */
            for (x1 = 0; x1 < sz; x1++)
            {
                /* the list index (1-40) of the second allele */
                for (x2 = 0; x2 < sz; x2++)
                {
                    big_mat->data[(y1 * sz + y2) * big_mat->size_x + x1 * sz + x2] = 
                        trans_mat.data[y1 * trans_mat.size_x + x1] * 
                        trans_mat.data[y2 * trans_mat.size_x + x2];
                }
            }
        }
    }
    
    /* build several state maps */
    
    /* full_letter has states such as 'ac/cg' i.e. the letterspace state 
     * 'ac/cg' would be given as [0 1 1 2]
     */
    maps->full_letter = va_allocMatU8(4, sz * sz);
    memset(maps->full_letter.data, 0, 
        maps->full_letter.size_x * maps->full_letter.size_y);

    for (i = 0; i < sz; i++)
    {
        int j;

        for (j = 0; j < sz; j++)
        {
            char ref[4];
            int k;
            int idx;
            VA_VecU8 r;
            VA_VecU8 p;

            ref[0] = list[i][0];
            ref[1] = list[i][1];
            ref[2] = list[j][0];
            ref[3] = list[j][1];

            r.data = ref;
            r.size = sizeof(ref);
 
            p = nuc2nuc(r);
            VA_ASSERT(p.data != NULL);

            for (k = 0; k < 4; k++)
            {
                idx = (i * sz + j) * maps->full_letter.size_x + k;
                maps->full_letter.data[idx] = p.data[k];
            }
            free(p.data);
       }
    }

    /* just the potential output letters in the state */

    maps->call_letter = va_allocMatU8(2, sz * sz);
    memset(maps->call_letter.data, 0, maps->call_letter.size_x *
        maps->call_letter.size_y);

    for (i = 0; i < sz; i++)
    {
        int j;

        for (j = 0; j < sz; j++)
        {
            int idx1 = (i * sz + j) * maps->call_letter.size_x;
            int idx2 = (i * sz + j) * maps->full_letter.size_x;

            maps->call_letter.data[idx1] = maps->full_letter.data[idx2 + 1];
            maps->call_letter.data[idx1 + 1] = maps->full_letter.data[idx2 + 3];
        }
    }

    /* the colors of the states */
    maps->call_color = va_allocMatU8(2, sz * sz);
    memset(maps->call_color.data, -1, maps->call_color.size_x *
        maps->call_color.size_y);
    {
        int color_1;
        size_t csz0 = sizeof(comb.c[0])/sizeof(comb.c[0][0]);
        size_t csz1 = sizeof(comb.c)/sizeof(comb.c[0]);
     
        for (color_1 = 0; color_1 < csz1; color_1++)
        {
            int pos_1;
            for (pos_1 = 0; pos_1 < csz0; pos_1++)
            {
                int color_2;
                for (color_2 = 0; color_2 < csz1; color_2++)
                {
                    int pos_2;
                    for (pos_2 = 0; pos_2 < csz0; pos_2++)
                    {    
                        int pos = (comb.c[color_1][pos_1] - 1) * sz + 
                                  comb.c[color_2][pos_2] - 1;
                        maps->call_color.data[pos * maps->call_color.size_x] = color_1 + 1;
                        maps->call_color.data[pos * maps->call_color.size_x + 1] = color_2 + 1;
                    }
                }
            }
        }
 
        for (i = 0; i < maps->call_color.size_x * maps->call_color.size_y; i++)
            VA_ASSERT(maps->call_color.data[i] != 0);
    }
    
    rc = clock_gettime(CLOCK_REALTIME, &te);
    VA_ASSERT(rc == 0);

    tdiff = va_TimeSub(&te, &ts);
    fprintf(stderr, "\t Done in: %ld.%03ld secs\n",
        tdiff.tv_sec, tdiff.tv_nsec / NS_IN_MS);
}    

