/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

/******************************************************************************
* VARiD run the VARiD algorithm
*   VARiD(cdata, ref_pos) run the cosin algorithm, where cdata is
*   filled as requested, and ref_pos is a dictionary of extra positions
*   and respective nucleotides to allow in the reference when
*   transitioning. More details are coming
*
*   Quick breakdown:
*       1) compute the transition breakdown, which determines what the
*       probability of transitioning from one state to another is.
*       2) compute the covmap and emission probability. The covmap computes
*       the coverages and some probabilities associates with the reads, and
*       then we compute the emission probability based on these coverages.
*       3) compute the forward and backward algorithm to get a final
*       probability.
*
*
*   TODO: verify code for probability matrix, normalization, and sanity
*   check
******************************************************************************/

static Comb_t comb = {
    {
        {1,   6,  11,  16,  25,  30,  35,  40},    /* color 1 */
        {2,   5,  12,  15,  26,  29,  36,  39},    /* color 2 */
        {3,   9,   8,  14,  27,  33,  32,  38},    /* color 3 */
        {4,   7,  10,  13,  28,  31,  34,  37},    /* color 4 */
        {17, 18,  19,  20,  21,  22,  23,  24}     /* gap */
    },
    {
        {1,   5,   9,  13,  25,  29,  33,  37},    /* letter a */
        {2,   6,  10,  14,  26,  30,  34,  38},    /* letter c */
        {3,   7,  11,  15,  27,  31,  35,  39},    /* letter g */
        {4,   8,  12,  16,  28,  32,  36,  40},    /* letter t */
        {17, 18,  19,  20,  21,  22,  23,  24}     /* no letter emission */
    }
};

void varid(
    CData_t        cdata, 
    VA_INT32       ref_pos,
    Maps_t       * maps,
    Covmaps_t    * covmap,
    VA_MatDouble * p,
    int useSam)
{
    struct timespec ts;
    struct timespec te;
    struct timespec tdiff;
    int rc;
    int i;
    int j;
    int n;
    Frequency_t freq;
    VA_MatDouble trans_mat;
    VA_MatDouble e;
    VA_MatDouble f;
    VA_MatDouble b;
    VA_VecDouble s_f;
    VA_VecDouble s_b;
    VA_VecDouble lsb;
    VA_VecDouble lsf;
    VA_VecDouble cumlsb;
    VA_VecDouble cumlsf;
    VA_VecI32 window;
    VA_DOUBLE Px_f;
    VA_DOUBLE Px_b;
    VA_DOUBLE slsf;
    VA_VecU8  skipf;
    VA_VecU8  skipb;
    VA_VecU8  skippos;

    /* initial time and printing */
    rc = clock_gettime(CLOCK_REALTIME, &ts);
    VA_ASSERT(rc == 0);

    fprintf(stderr, "Starting VARiD Algorithm \n");
    
    /* prepare data */
    
    /*
     * we will work with a 40*gen_length matrix (there are 40 possible
     * states for each position). the comb stuct indicates
     * which states each color/letter is present in: for example, we have
     * color 0 in states 1, 6, 11, 16, ...
     *
     * This is because the states are done as follows: (reminder: capital L
     * is a gap with L memory, see get_transitions function)
     *   states 01-16: {a,c,g,t}x{a,c,g,t}={aa, ac, ag, at, ca, ..., tg, tt}
     *   states 17:20: {aA, cC, gG, tT}
     *   states 20:24: {AA, CC, GG, TT}
     *   states 25:40: {A,C,G,T}x{a,c,g,t}={Aa, Ac, Ag, At, Ca, ..., Tg, Tt}
     */
    /* color emissions */
    /* emit color 1:4 (really colors 0:3), and c(5) = gap emission */
    for (i = 0; i < sizeof(comb.c) / sizeof(comb.c[0]); i++)
         for (j = 0; j < sizeof(comb.c[0])/sizeof(comb.c[0][0]); j++)
             VA_ASSERT(comb.c[i][j] != 0);

    for (i = 0; i < sizeof(comb.l) / sizeof(comb.l[0]); i++)
         for (j = 0; j < sizeof(comb.l[0])/sizeof(comb.l[0][0]); j++)
             VA_ASSERT(comb.l[i][j] != 0);
    
    /* frequencies */
    /* set the frequencies of the various letters in the reference */

    memset(&freq, 0, sizeof(freq));

    for (i = 0; i < cdata.ref.size; i++)
    {
        freq.a += (cdata.ref.data[i] == 0) ? 1.0 : 0.0; 
        freq.c += (cdata.ref.data[i] == 1) ? 1.0 : 0.0; 
        freq.g += (cdata.ref.data[i] == 2) ? 1.0 : 0.0; 
        freq.t += (cdata.ref.data[i] == 3) ? 1.0 : 0.0; 
    }

    freq.a = freq.a / cdata.ref.size;
    freq.c = freq.c / cdata.ref.size;
    freq.g = freq.g / cdata.ref.size;
    freq.t = freq.t / cdata.ref.size;

    /* set the frequenciesof various letters in the reference */
    freq.gapopen = 0.001;
    freq.gapextend = 0.001;

    freq.a = 1.0/4 - freq.gapopen / 4;
    freq.c = 1.0/4 - freq.gapopen / 4;
    freq.g = 1.0/4 - freq.gapopen / 4;
    freq.t = 1.0/4 - freq.gapopen / 4;

    /* 
     * set the gap open and gap extension penalties.
     * TODO: should these be smaller? should they be provided in cdata?
     */
    freq.snp_penalty = cdata.snp_rate;

    /* compute the transitions matrices */

    /* computation */
    memset(&trans_mat, 0, sizeof(trans_mat));
    get_transitions(freq, comb, &trans_mat, maps);

    /* compute the coverage maps */
    *covmap = varid_covmaps(cdata, useSam);
    
    /* compute emissions */
    e = varid_emissions(cdata, trans_mat, *covmap, comb);
    
    /* run forward and backward algorithm */
    window = va_allocVecI32(2);
    window.data[0] = 1;
    window.data[1] = cdata.ref.size;

    Px_f = varid_forward(trans_mat, cdata, e, *maps, window, &f, &skipf, &s_f);
    Px_b = varid_backward(trans_mat, cdata, e, *maps, window, &b, &skipb, &s_b);

    skippos = va_allocVecU8(skipf.size);
    for (i = 0; i < skippos.size; i++)
        skippos.data[i] = skipf.data[i] | skipb.data[i];

    /* form the probability matrix P */
    *p = va_allocMatDouble(f.size_x, f.size_y);
    memset(p->data, 0, sizeof(VA_DOUBLE) * p->size_x * p->size_y);

    lsf = va_allocVecDouble(s_f.size);
    cumlsf = va_allocVecDouble(s_f.size);

    lsf.data[0] = log(s_f.data[0]);
    slsf = lsf.data[0];
    cumlsf.data[0] = lsf.data[0];
  
    for (i = 1; i < s_f.size; i++)
    {
        lsf.data[i] = log(s_f.data[i]);
        slsf += lsf.data[i];
        cumlsf.data[i] = lsf.data[i] + cumlsf.data[i - 1];
    }

    lsb = va_allocVecDouble(s_b.size);
    cumlsb = va_allocVecDouble(s_b.size);
    for (i = 0; i < s_b.size; i++)
        lsb.data[i] = log(s_b.data[i]);

    fliplrDouble(lsb);
    cumlsb.data[0] = lsb.data[0];

    for (i = 1; i < s_b.size; i++)
        cumlsb.data[i] = lsb.data[i] + cumlsb.data[i - 1];
    fliplrDouble(cumlsb);

    n = 0;
    for (i = 0; i < s_f.size; i++)
    {
        int k;

        /* faster computation */
        for (k = 0; k < p->size_y; k++)
        {
            p->data[k * p->size_x + i] = exp(cumlsf.data[i] + cumlsb.data[i] - slsf + 
                                             log(f.data[k * f.size_x + i]) +
                                             log(b.data[k * b.size_x + i]));
        }
    }

    free(lsb.data);
    free(lsf.data);
    free(cumlsb.data);
    free(cumlsf.data);
    
    /* compute each part in turn if had to split the regions ... */
    for (i = 0; i < skippos.size; i++)
        if (skippos.data[i] == TRUE)
            n++;
       
    if (n > 0)
    {
        VA_VecI32 skiposf;
        int x;
        int rngstart;
        int rngend;
        int nskipf;
        int nskipb; 
        VA_DOUBLE pmax;

        free(p->data);
        skiposf = va_allocVecI32(n+2);
        n = 0;
        skiposf.data[n++] = 1;
        for (i = 0; i < skippos.size; i++)
            if (skippos.data[i] == TRUE)
                skiposf.data[n++] = i + 1;
        skiposf.data[n++] = cdata.ref.size;

        *p = va_allocMatDouble(f.size_x, f.size_y);
        memset(p->data, 0, sizeof(VA_DOUBLE) * p->size_x * p->size_y);

        for (x = 1; x <= skiposf.size - 1; x++)
        {
            VA_MatDouble f_s;
            VA_MatDouble b_s;
            VA_DOUBLE sum_sf;
            VA_DOUBLE sum_sb;

            rngstart = skiposf.data[x - 1];
            rngend = skiposf.data[x] - 1;
            if (skiposf.data[x] == skiposf.data[skiposf.size - 1])
                rngend = skiposf.data[skiposf.size - 1];
            window.data[0] = rngstart + 1;
            window.data[1] = rngend - 1;

            if (window.data[1] - window.data[0] < 4)
            {
                fprintf(stderr, "\t\tSkipping extra-small window: [%d %d]\n",
                    rngstart, rngend);
                continue;
            }

            fprintf(stderr, "\t\tRunning FB for window: [%d %d]\n",
                    rngstart, rngend);

            Px_f = varid_forward(trans_mat, cdata, e, *maps, 
                                 window, &f_s, &skipf, &s_f);
            Px_b = varid_backward(trans_mat, cdata, e, *maps, 
                                  window, &b_s, &skipb, &s_b);

            nskipf = FALSE;
            for (i = 0; (nskipf == FALSE) && (i < skipf.size); i++)
                if (skipf.data[i] == TRUE)
                    nskipf = TRUE;
            nskipb = FALSE; 
            for (i = 1; (nskipb == FALSE) && (i < skipb.size); i++)
                if (skipb.data[i] == TRUE)
                    nskipb = TRUE;

            if (nskipf || nskipb)
                fprintf(stderr, "WARNING: DETECTED re-skip in window\n");
        
            sum_sf = 0.0;
            for (i = 0; i < s_f.size; i++)
               sum_sf += log(s_f.data[i]);
 
            sum_sb = 0.0;
            for (i = 0; i < s_b.size; i++)
               sum_sb += log(s_b.data[i]);
 
            fprintf(stderr, "\t\tlogPxf:%.4e, logPxb:%.4e\n", sum_sf, sum_sb);

            for (i = 0; i < s_f.size; i++)
            {
                VA_DOUBLE d1 = 0.0;
                VA_DOUBLE d2 = 0.0;
                VA_DOUBLE d3 = 0.0;
                VA_DOUBLE d;
                int k;

                for (k = 0; k <= i; k++)
                    d1 += log(s_f.data[k]);
                for (k = i; k < s_b.size; k++)
                    d2 += log(s_b.data[k]);
                for (k = 0; k < s_f.size; k++)
                    d3 += log(s_f.data[k]);

                for (k = 0; k < p->size_y; k++)
                {
                    d = d1 + d2 - d3 +
                        log(f_s.data[k * f_s.size_x + i]) +
                        log(b_s.data[k * b_s.size_x + i]);

                    if (!isinf(d))
                        p->data[k * p->size_x + window.data[0] - 1 + i] = exp(d);
                }
            }

            for (i = window.data[0]; i <= window.data[1]; i++)
            {
                f.data[i - 1] = f_s.data[i - window.data[0]];
                b.data[i - 1] = b_s.data[i - window.data[0]];
            }

            pmax = p->data[window.data[0] - 1];
            for (i = 0; i < p->size_y; i++)
            {
                for (j = window.data[0]; j <= window.data[1]; j++)
                    pmax = MAX(pmax, p->data[i * p->size_x + j - 1]);
            }

            fprintf(stderr, "\t\t max p in this region %.4e\n", pmax); 
            free(f_s.data);
            free(b_s.data);
        }
    }

    free(skippos.data);
    free(window.data);
    free(e.data);
    free(trans_mat.data);
 
    /* print time results */
    rc = clock_gettime(CLOCK_REALTIME, &te);
    VA_ASSERT(rc == 0);

    tdiff = va_TimeSub(&te, &ts);

    /* done read prep */
    fprintf(stderr, "\t DONE VARiD Algorithm & prediction in: %ld.%03ld secs\n",
            tdiff.tv_sec, tdiff.tv_nsec/NS_IN_MS);
}    
