/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */
 
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

VA_Vector t_sparse;
/******************************************************************************
*   Here we compute the Backward part of the Forward-Backward HMM Algorithm.
*
*   Author: Adrian V. Dalca
*******************************************************************************/
VA_DOUBLE varid_backward(
    VA_MatDouble   trans_mat, 
    CData_t        cdata, 
    VA_MatDouble   e, 
    Maps_t         maps, 
    VA_VecI32      window,
    VA_MatDouble * b, 
    VA_VecU8     * skip_pos, 
    VA_VecDouble * s)
{
    struct timespec ts;
    struct timespec te;
    struct timespec tdiff;
    int rc;
    int i;
    int j;
    int k;
    VA_VecDouble tmat2_use;
    VA_VecDouble inmat;
    VA_MatU8 o;
    VA_VecU8 isgap;
    VA_VecU8 snps1;
    VA_VecU8 snps2;
    size_t sz = cdata.ref.size;
    VA_DOUBLE Px_b;
    VA_MatDouble b2;
    VA_VecDouble s2;
    int n1;
    int n2;
    VA_DOUBLE eps = 2.2204e-016;
    VA_Vector * p_mats = NULL;
    int t_sparse_csz = 0;
    SparseDNode_t * sp;


    VA_ASSERT(b != NULL);
    VA_ASSERT(s != NULL);
    VA_ASSERT(skip_pos != NULL);

    rc = clock_gettime(CLOCK_REALTIME, &ts);
    VA_ASSERT(rc == 0);

    /* initiate backward probability matrix */

    t_sparse.data = (SparseDNode_t *)malloc(sizeof(SparseDNode_t) * trans_mat.size_x);
    VA_ASSERT(t_sparse.data != NULL);
    t_sparse.size = trans_mat.size_x;
    t_sparse_csz = 0;

    for (i = 0; i < trans_mat.size_y; i++)
    {
        for (j = 0; j < trans_mat.size_x; j++)
        {
            if (trans_mat.data[i * trans_mat.size_x + j] != 0.0)
            {
                if (t_sparse_csz >= t_sparse.size)
                {
                    t_sparse.size *= 10;
                    t_sparse.data = (SparseDNode_t *)realloc(t_sparse.data,
                                     sizeof(SparseDNode_t) * t_sparse.size);
                    VA_ASSERT(t_sparse.data != NULL);
                }

                sp = ((SparseDNode_t *)t_sparse.data);
                sp[t_sparse_csz].x = j;
                sp[t_sparse_csz].y = i;
                sp[t_sparse_csz++].val = trans_mat.data[i * trans_mat.size_x + j];
            }
        }
    }

    t_sparse.data = (SparseDNode_t *)realloc(t_sparse.data,
                     sizeof(SparseDNode_t) * t_sparse_csz);
    VA_ASSERT(t_sparse.data != NULL);
    t_sparse.size = t_sparse_csz;

    sp = ((SparseDNode_t *)t_sparse.data);

    isgap = va_allocVecU8(maps.call_letter.size_y);
    for (i = 0; i < isgap.size; i++)
    {
        int csz_x = maps.call_letter.size_x;
        isgap.data[i]= ((maps.call_letter.data[i * csz_x] == 4) ||
                       (maps.call_letter.data[i * csz_x + 1] == 4)) ? TRUE : FALSE;
    }
    snps1 = va_allocVecU8(maps.call_letter.size_y);
    snps2 = va_allocVecU8(maps.call_letter.size_y);

    o = va_allocMatU8(trans_mat.size_x, trans_mat.size_y);

    /*
     * prepare transition matrices by not allowing more than a
     * single SNP at a time...
     */
    p_mats = (VA_Vector *)malloc(sizeof(VA_Vector) * 4 * 4);
    VA_ASSERT(p_mats != NULL);

    for (i = 0; i < 4; i++)
    {
        for (j = 0; j < 4; j++)
        {
            p_mats[j * 4 + i].data = 
                 (SparseDNode_t *)malloc(t_sparse.size * sizeof(SparseDNode_t));
            p_mats[j * 4 + i].size = t_sparse.size;
        }
    }

    for (n1 = 0; n1 <= 3; n1++)
    {
        int lsz = maps.call_letter.size_x;

        /*
         * get all the non-gap states where at least one of the alleles is
         * NOT letter1 (which will be the reference)
         */
        for (j = 0; j < snps1.size; j++)
            snps1.data[j] = (((maps.call_letter.data[j * lsz] != n1) ||
                 (maps.call_letter.data[j * lsz + 1] != n1)) &&
                 (isgap.data[j] == FALSE)) ? TRUE : FALSE;
        for (n2 = 0; n2 <= 3; n2++)
        {
            int i_curr = 0;
            /* get all the non-gap states where at least one of the alleles
             * is NOT letter2 (which will be the reference)
             */
            VA_Vector * pm_sparse = &p_mats[n2 * 4 + n1];

            for (j = 0; j < snps2.size; j++)
                snps2.data[j] = (((maps.call_letter.data[j * lsz] != n2) ||
                     (maps.call_letter.data[j * lsz + 1] != n2)) &&
                     (isgap.data[j] == FALSE)) ? TRUE : FALSE;

            /*
             * now, eradicate any chance of going from snps1 to snps2
             * states, i.e. do not allow adjacent snps. This rule can be
             * avoided IF the reference allows multiple alleles in
             * subsequent runs.
             */
            for (i = 0; i < o.size_y; i++)
            {
                int y = i * o.size_x;

                for (j = 0; j < o.size_x; j++)
                {
                    if ((snps2.data[j] == TRUE) && (snps1.data[i] == TRUE))
                        o.data[y + j] = 0;
                    else
                        o.data[y + j] = 1;
                }
            }

            sp = (SparseDNode_t *)t_sparse.data;
            for (i = 0; i < pm_sparse->size; i++)
            {
                if (o.data[sp[i].y * o.size_x + sp[i].x] > 0)
                {
                    ((SparseDNode_t *)pm_sparse->data)[i_curr].x = sp[i].x;
                    ((SparseDNode_t *)pm_sparse->data)[i_curr].y = sp[i].y;
                    ((SparseDNode_t *)pm_sparse->data)[i_curr++].val = sp[i].val;
                }
            }

            if (i_curr != t_sparse.size)
            {
                pm_sparse->data = (SparseDNode_t *)realloc(pm_sparse->data,
                                         i_curr * sizeof(SparseDNode_t));
                VA_ASSERT(pm_sparse->data != NULL);
                pm_sparse->size = i_curr;
            }
        }
    }

    free(o.data);
    free(snps1.data);
    free(snps2.data);
    free(isgap.data);

    /********************************************************************/
    /* BACKWARD ALGORITHM                                               */
    /********************************************************************/

    tmat2_use = va_allocVecDouble(trans_mat.size_x);

    /* initiate backward probability matrix (b) and normalization (s) */
    *s = va_allocVecDouble(sz);
    memset(s->data, 0, sizeof(VA_DOUBLE) * s->size);

    *b = va_allocMatDouble(sz, trans_mat.size_y);
    memset(b->data, 0, sizeof(VA_DOUBLE) * b->size_x * b->size_y);

    /* initiate - TODO: MAYBE WHOULD INITIATE TO THE LATEST CHREND? */
    s->data[window.data[1] - 1] = trans_mat.size_y;

    for (i = 0; i < b->size_y; i++)
        b->data[i * b->size_x + window.data[1] - 1] = 
            1.0 / s->data[window.data[1] - 1];

    *skip_pos = va_allocVecU8(sz);
    for (i = 0; i < skip_pos->size; i++)
        skip_pos->data[i] = FALSE;

    inmat = va_allocVecDouble(e.size_y);
 
    /* backward algorithm through the positions of the genome */
    for (k = window.data[1] - 1; k >= window.data[0]; k--) 
    {
        VA_Vector * pm_sparse;
        SparseDNode_t * sp;

        /*
         * if both the previous and current positions in the reference are
         * actual nucleotides (and not N or gaps), use one of the prepared
         * transition matrices that doesn't allow adjacent gaps.
         */
        int prev_ref = cdata.ref.data[k - 1];
        int cur_ref = cdata.ref.data[k];

        memset(tmat2_use.data, 0, sizeof(VA_DOUBLE) * tmat2_use.size);
        if ((cur_ref >= 0) && (cur_ref <= 3)
            && (prev_ref >= 0) && (prev_ref <= 3))
        {
            pm_sparse = &p_mats[cur_ref * 4 + prev_ref];
        }
        else
        {
            pm_sparse = &t_sparse;
        }

        /* compute the backwards probability (m), unnomalized. */
        for (i = 0; i < e.size_y; i++)
            inmat.data[i] = e.data[i * e.size_x + k] * b->data[i * b->size_x + k];

        sp = (SparseDNode_t *)pm_sparse->data;

        for (i = 0; i < pm_sparse->size; i++)
            tmat2_use.data[sp[i].y] += sp[i].val * inmat.data[sp[i].x];

        /* normalize the backwards probability b(i) */ 
        s->data[k - 1] = 0.0;
        for (i = 0; i < tmat2_use.size; i++)
            s->data[k - 1] += tmat2_use.data[i]; 

       
        if (s->data[k - 1] < eps)
        {
            s->data[k - 1] = trans_mat.size_y;
            fprintf(stderr, "Resetting backward s, position: %d\n", k);
            skip_pos->data[k - 1] = TRUE;
            
            for (i = 0; i < b->size_y; i++)
                b->data[i * b->size_x + k - 1] = 1.0 / s->data[k - 1];
        }
        else
        {
            for (i = 0; i < b->size_y; i++)
                b->data[i * b->size_x + k - 1] = tmat2_use.data[i] / s->data[k - 1];
        }
    }
    
    /* do i need to initiate with [1, 0, 0, 0...] here to make this work? */
    s2 = va_allocVecDouble(window.data[1] - window.data[0] + 1);
    memcpy(s2.data, &s->data[window.data[0] - 1], 
           (window.data[1] - window.data[0] + 1) * sizeof(VA_DOUBLE));
    free(s->data);
    memcpy(s, &s2, sizeof(s2));

    b2 = va_allocMatDouble(window.data[1] - window.data[0] + 1, b->size_y);
    for (i = 0; i < b2.size_y; i++)
        memcpy(&b2.data[i * b2.size_x], 
               &b->data[i * b->size_x + window.data[0] - 1],
               (window.data[1] - window.data[0] + 1) * sizeof(VA_DOUBLE));
    free(b->data);
    memcpy(b, &b2, sizeof(b2));

    Px_b = 0.0;
    for (i = 0; i < s->size; i++)
        Px_b += log(s->data[i]);

    rc = clock_gettime(CLOCK_REALTIME, &te);
    VA_ASSERT(rc == 0);

    tdiff = va_TimeSub(&te, &ts);
    fprintf(stderr, "\t\tbwd algo done in: %ld.%03ld secs\n",
        tdiff.tv_sec, tdiff.tv_nsec/NS_IN_MS);

    if (inmat.data != NULL)
        free(inmat.data);
    if (tmat2_use.data != NULL)
        free(tmat2_use.data);

    free(t_sparse.data);

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            free(p_mats[j * 4 + i].data);
    free(p_mats);

    return (Px_b);
}
