/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <math.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

/******************************************************************************
* VARID_COVMAPS computes the letter and colorspace coverage maps
*   Here we compute a few coverage-related stats. We compute the overall
*   coverage, the number of times we see each color and each letter, and
*   finally the emission support for a color (col_em) and against a
*   color (noncol_em), and the same for letters. More details to come.
*
*
*
*
*   GEN_SIZE, READSTARTS, READS
*
*
*   note: stats CUTOFFs, size CUTOFFs
* TODO: quality values
* !!! Asuuming T start when we do first letter translation???
*******************************************************************************/

extern double*  cs_error_cnt;
extern double*  cs_error_dpts;
extern unsigned int cs_error_size;
extern double*  ls_error_cnt;
extern double*  ls_error_dpts;
extern unsigned int ls_error_size;



Covmaps_t varid_covmaps(
    CData_t cdata,
    int useSam)
{
    struct timespec ts;
    struct timespec te;
    struct timespec tdiff;
    int rc;
    Covmaps_t covmap;
    int usedreads = 0; /* the number of reads used */
    int i;
    int j;
    size_t sz = cdata.ref.size;
    size_t nrreads = cdata.reads.size;
    ReadsNode_t * rd = (ReadsNode_t *)cdata.reads.data;
    int zeroindexskippedreads = 0;

    rc = clock_gettime(CLOCK_REALTIME, &ts);
    VA_ASSERT(rc == 0);

    /* coverage maps of various kind */
    /* the amount of coverage over a certain position from each allele */
    covmap.coverage = va_allocMatI32(sz, 2);  
    memset(covmap.coverage.data, 0, sizeof(VA_INT32) * 
           covmap.coverage.size_x * covmap.coverage.size_y);

    /* color emissions, in log space */
    /* histogram of colors */
    covmap.colors = va_allocMatDouble(sz, 5);  
    memset(covmap.colors.data, 0, sizeof(VA_DOUBLE) * 
           covmap.colors.size_x * covmap.colors.size_y);

    /* col_emission */
    covmap.col_em = va_allocMatDouble(sz, 5);  
    memset(covmap.col_em.data, 0, sizeof(VA_DOUBLE) * 
           covmap.col_em.size_x * covmap.col_em.size_y);

    /* noncol_emission */
    covmap.noncol_em = va_allocMatDouble(sz, 5);  
    memset(covmap.noncol_em.data, 0, sizeof(VA_DOUBLE) * 
           covmap.noncol_em.size_x * covmap.noncol_em.size_y);

    /* letter emissions, in log space */
    /* histogram of letters */
    covmap.letters = va_allocMatDouble(sz, 5);  
    memset(covmap.letters.data, 0, sizeof(VA_DOUBLE) * 
           covmap.letters.size_x * covmap.letters.size_y);

    /* let_emission */
    covmap.let_em = va_allocMatDouble(sz, 5);  
    memset(covmap.let_em.data, 0, sizeof(VA_DOUBLE) * 
           covmap.let_em.size_x * covmap.let_em.size_y);

    /* nonlet_emission */
    covmap.nonlet_em = va_allocMatDouble(sz, 5);  
    memset(covmap.nonlet_em.data, 0, sizeof(VA_DOUBLE) * 
           covmap.nonlet_em.size_x * covmap.nonlet_em.size_y);

    for (i = 0; i < nrreads; i++)
    {
        VA_VecDouble ferr;
        VA_VecDouble err;
        VA_VecDouble logerr;
        VA_VecDouble logcor;
        VA_VecU8 nongaps;
        VA_VecU8 indices;
        VA_VecU8 ind;
        VA_VecI32 seqrange;
        int isLetterspace;
        int isForward;
        int haplome;
        int readlen;
        int rangestart;
        int rangeend;
        int errorstart;
        int nrnongaps;
        int szseq;
        int csCond;
        int lsCond;
        VA_VecDouble error_mat;
        int cut;
        int runoff;
        VA_VecI32 allind;
        int k;
       
        memset(&seqrange, 0, sizeof(seqrange));
        /* verification and setup some useful variables */

        /* update the number of reads */
        isForward = (rd[i].strand == '+') ? TRUE : FALSE;
        isLetterspace = (rd[i].seqtype == LETTERSPACE_READ) ? TRUE: FALSE;

        /* only count reads that have at least a certain norm odds and
         * pgenome
         */
#define MAPQ_CUTOFF		(0)
        if (useSam){
        	lsCond = rd[i].extras.nodds < MAPQ_CUTOFF;
        	csCond = lsCond;
        }else{
        csCond = ((rd[i].extras.nodds < cdata.normodds_cutoff) ||
        		(rd[i].extras.pgenome < cdata.pgenome_cutoff)) ? TRUE : FALSE;
        lsCond = ((rd[i].extras.nodds < cdata.normodds_cutoff/2) ||
        		(rd[i].extras.pgenome < cdata.pgenome_cutoff)) ? TRUE : FALSE;
        }
        if ((isLetterspace && lsCond) || (!isLetterspace && csCond))
        {        
            continue;
        }
        
        /* assign the haplome */
        if (cdata.src == SRC_REAL)
        {
            /* is real data, we don't know the haplome, so we work with 1 by
             * default
             */
            haplome = 1;
        }
        else
        {
            haplome = rd[i].extras.haplome;
        }

        if (isLetterspace)
        {
            readlen = MIN(rd[i].letseq.size, ls_error_size);
            readlen = MIN(readlen,  cdata.readsize_cutoff_ls);
            nongaps = va_allocVecU8(rd[i].letseq.size);
            for (j = 0; j < rd[i].letseq.size; j++)
                nongaps.data[j] = (rd[i].letseq.data[j] != 4) ? 1 : 0;

            /* TODO: assuming a 454 error of 4%. This way be too high, but
             * it is the cited per-position error of Roche GS20.
             * Ref: Huse, S.M et al, 2007
             */
             error_mat = va_allocVecDouble(readlen);
             for (j = 0; j < error_mat.size; j++)
                 error_mat.data[j] = cdata.error_mat_ls[j];
        }
        else
        {
            readlen = MIN(rd[i].colseq.size, cs_error_size);
	    readlen = MIN(readlen,  cdata.readsize_cutoff_cs);
            nongaps = va_allocVecU8(rd[i].colseq.size);
            for (j = 0; j < rd[i].colseq.size; j++)
                nongaps.data[j] = (rd[i].colseq.data[j] != 4) ? 1 : 0;

             error_mat = va_allocVecDouble(sizeof(cdata.error_mat_cs)/sizeof(cdata.error_mat_cs[0]));
             for (j = 0; j < error_mat.size; j++)
                 error_mat.data[j] = cdata.error_mat_cs[j];
        }

        /* prepare the ranges that this read covers the haplome over.
         * seqrange - only the range of the sequence - this will be the same
         * as the full range for letterspace sequences. For colorspace, it
         * will be one less.
         * We also perform a read cutoff - in general with NGS platforms we
         * notice a drop in quality towards the end of the reads. It is
         * therefore useful to sometimes cut the ends of the reads off.
         * TODO: This can be cleaned up a little, maybe.
         */
        if (isLetterspace && isForward)
        {
            rangestart = rd[i].chrstart;
            rangeend = rangestart + readlen - 1;
            errorstart = 1;
        }
        else if (isLetterspace && !isForward)
        {
            rangeend = rd[i].chrend;
            rangestart = rangeend - readlen + 1;
            errorstart = 1; /* but backwards */
        }
        else if (!isLetterspace && isForward)
        {
            /* start a little later since we translate the first letter.
             * That letter will be treated separately from this range.
             */
            rangestart = rd[i].chrstart + 1;
            rangeend = rangestart + readlen - 2;
            errorstart = 2;
        }
        else if (!isLetterspace && !isForward)
        {
            /* backward colorspace */
            rangeend = rd[i].chrend;
            rangestart = rangeend - readlen + 2;
            errorstart = 2; /* but backwards */
        }

        szseq = MIN(rangeend, sz) - rangestart + 1;
        if (szseq > 0)
        {
            seqrange = va_allocVecI32(szseq);
            for (j = 0; j < seqrange.size; j++)
                seqrange.data[j] = rangestart + j;
        }
        else
        {
            continue;
        }

        cut = rangeend - seqrange.data[seqrange.size - 1];

        /* quick test to see if */
        runoff = 0;
        if (cdata.ref.size < rangeend)
            runoff = rangeend - cdata.ref.size;
          
        /* here we correct the error vector. We should have a master error
         * vector that contains the error distribution for our reads. next,
         * we take only the actual reads (ignoring gaps) and extract that
         * length from the error vector. (starting with the first entry in
         * the case of letterspace, or second entry in the case of
         * colorspace, as determined by 'errorstart').
         */
        nrnongaps = 0;
        for (j = errorstart; j <= errorstart + seqrange.size - 1; j++)
            nrnongaps += nongaps.data[j - 1]; 

        if (nrnongaps > 0)
           ferr = va_allocVecDouble(nrnongaps);
        else
           memset(&ferr, 0, sizeof(ferr));
   
        for (j = errorstart; j <= errorstart + nrnongaps - 1; j++)
            ferr.data[j - errorstart] = error_mat.data[j - 1];
		//fprintf(stderr,"1:%p\n",error_mat.data);
        free(error_mat.data);

        if (!isForward)
        {
           if (nrnongaps > 0)
               fliplrDouble(ferr);
        }

        err = va_allocVecDouble(seqrange.size);
        for (j = 0; j < err.size; j++)
            err.data[j] = cdata.gap_error; 

        k = 0;
        for (j = errorstart; j <= errorstart + seqrange.size - 1; j++)
            if ((nongaps.data[j - 1] == 1) && (k < ferr.size))
                err.data[j - errorstart] = ferr.data[k++];
        

        /* we now compute the emission probability in case of error and in
         * case of a correct call
         */
        logerr = va_allocVecDouble(err.size);
        for (j = 0; j < logerr.size; j++)
            logerr.data[j] = log(err.data[j] / 3); /* log emission of an error */

        logcor = va_allocVecDouble(err.size);
        for (j = 0; j < logerr.size; j++)
            logcor.data[j] = log(1 - err.data[j]);  /* log emission of a correct call. */

        /**************************************************************/
        /* Update covmaps                                             */
        /**************************************************************/

        /* if letterspace, only update the letterspace coverage */
        if (isLetterspace)
        {
            VA_VecU8 indices2;
            VA_VecI32 seqrange2;
            size_t uch = 0;

            /* note: "+1" usage in indexing due to MATLAB indexing format:
             * an observed color 0 would need to be put in index 1.
             */
            indices = va_allocVecU8(seqrange.size);
            if (isForward)
            {
                memcpy(indices.data, rd[i].letseq.data, 
                    indices.size * sizeof(VA_UINT8));
            }
            else
            {
                size_t szfrom = rd[i].letseq.size - cut - seqrange.size;

                memcpy(indices.data, &rd[i].letseq.data[szfrom], 
                    indices.size * sizeof(VA_UINT8));
            }

                 
            /* find any unknown characters */
            for (j = 0; j < indices.size; j++)
                if (indices.data[j] == 5)
                   uch++;

            if (uch > 0)
            {
                indices2 = va_allocVecU8(indices.size - uch);
                seqrange2 = va_allocVecI32(seqrange.size - uch);
                indices2.size = 0;
                seqrange2.size = 0;

                for (j = 0; j < indices.size; j++)
                {
                    if (indices.data[j] != 5)
                    {
                        indices2.data[indices2.size++] = indices.data[j];
                        seqrange2.data[seqrange2.size++] = seqrange.data[j];
                    }
                }
				//fprintf(stderr,"2:%p\n",indices.data);
                free(indices.data);
                //fprintf(stderr,"3:%p\n",seqrange.data);
                free(seqrange.data);
                memcpy(&indices, &indices2, sizeof(indices2));
                memcpy(&seqrange, &seqrange2, sizeof(seqrange2));
            } 

            if (indices.size == 0)
            {
                zeroindexskippedreads++;
                continue;
            }

            allind = va_allocVecI32(seqrange.size);
            for (j = 0; j < indices.size; j++)
            {
                allind.data[j] = covmap.letters.size_y * (seqrange.data[j] - 1) + 
                                 indices.data[j] + 1;
            }

            for (j = 0; j < allind.size; j++)
            {
                size_t x = allind.data[j] / covmap.letters.size_y;
                size_t y = allind.data[j] - x * covmap.letters.size_y;
                size_t xy;

                if (y == 0)
                {
                    y = covmap.letters.size_y;
                    VA_ASSERT(x > 0);
                    x--;
                }
                
                xy = (y - 1) * covmap.letters.size_x + x;

                covmap.letters.data[xy]++;
                covmap.let_em.data[xy] += 
                    logcor.data[seqrange.data[j] - seqrange.data[0]];
                covmap.nonlet_em.data[xy] += 
                    logerr.data[seqrange.data[j] - seqrange.data[0]];
            }
            //fprintf(stderr,"4:%p\n",allind.data);
            free(allind.data);
            //fprintf(stderr,"5:%p\n",indices.data);
            free(indices.data);
        }
        else
        {
            int pos;
            int useQV;
            VA_VecU8 onoff;

            if (seqrange.size == 0)
            {
                zeroindexskippedreads++;
                continue;
            }
            /* otherwise, for colorspace update both the colorspace coverage
             * and the translated-through letter
             */
            indices = va_allocVecU8(seqrange.size);
            onoff = va_allocVecU8(indices.size);
            
            if (isForward)
            {
                memcpy(indices.data, rd[i].colseq.data,
                    indices.size * sizeof(VA_UINT8));
                for (j = 0; j < onoff.size; j++)
                    onoff.data[j] = (rd[i].qvals.data[1 + j] >= 10) ? 1 : 0;
            }
            else
            {
                size_t szfrom = rd[i].colseq.size - runoff - seqrange.size;

                VA_ASSERT(szfrom >= 0);
                memcpy(indices.data, &rd[i].colseq.data[szfrom],
                    indices.size * sizeof(VA_UINT8));
                for (j = 0; j < onoff.size; j++)
                    onoff.data[j] = (rd[i].qvals.data[1 + j] >= 10) ? 1 : 0;
                fliplrU8(onoff);
            }

            useQV = TRUE;
            if (useQV == FALSE)
            {
                for (j = 0; j < onoff.size; j++)
                   onoff.data[j] = 1;
            }

            allind = va_allocVecI32(seqrange.size);
            for (j = 0; j < indices.size; j++)
            {
                allind.data[j] = covmap.colors.size_y * (seqrange.data[j] - 1) + 
                                 indices.data[j] + 1;
            }

            for (j = 0; j < allind.size; j++)
            {
                size_t x = allind.data[j] / covmap.colors.size_y;
                size_t y = allind.data[j] - x * covmap.colors.size_y;
                size_t xy;
  
                if (y == 0)
                {
                    y = covmap.colors.size_y;
                    VA_ASSERT(x > 0);
                    x--;
                }

                xy = (y - 1) * covmap.colors.size_x + x;
                covmap.colors.data[xy] += onoff.data[j];
                covmap.col_em.data[xy] += 
                    onoff.data[j] * logcor.data[seqrange.data[j] - seqrange.data[0]];
                covmap.col_em.data[xy] += logcor.data[seqrange.data[j] - seqrange.data[0]];
                covmap.noncol_em.data[xy] += 
                    onoff.data[j] * logerr.data[seqrange.data[j] - seqrange.data[0]];
            }
			//fprintf(stderr,"6:%p\n",allind.data);
            free(allind.data);
            //fprintf(stderr,"7:%p\n",indices.data);
            free(indices.data);
            //fprintf(stderr,"8:%p\n",onoff.data);
            free(onoff.data);

            if (runoff == 0)
            {
                /* for the translated - through letters */
                ind = va_reallocVecU8(rd[i].letseq, rd[i].letseq.size);
                VA_ASSERT(ind.size == 1);

                if (isForward)
                    pos = seqrange.data[0] - 1;
                else
                    pos = seqrange.data[seqrange.size - 1];

                VA_ASSERT(ind.data[0] * covmap.letters.size_x + pos - 1 < 
                               covmap.letters.size_y *
                               covmap.letters.size_x);
                VA_ASSERT(ind.data[0] * covmap.let_em.size_x + pos - 1 < 
                               covmap.let_em.size_y *
                               covmap.let_em.size_x);
                VA_ASSERT(ind.data[0] * covmap.nonlet_em.size_x + pos - 1 <
                               covmap.nonlet_em.size_y *
                               covmap.nonlet_em.size_x);
                covmap.letters.data[ind.data[0] * covmap.letters.size_x + pos - 1] += 1;
                covmap.let_em.data[ind.data[0] * covmap.let_em.size_x + pos - 1] += 
                    log(1 - cdata.error_mat_cs[0]);
                covmap.nonlet_em.data[ind.data[0] * covmap.nonlet_em.size_x + pos - 1] +=
                    log(cdata.error_mat_cs[0] / 3);
            }
        }

        /* update the overall coverage */
        for (j = 1; j <= seqrange.size; j++)
        {
            int idx = (haplome - 1) * covmap.coverage.size_x +
                      seqrange.data[j - 1] - 1;
            VA_ASSERT(idx < covmap.coverage.size_x * covmap.coverage.size_y);
            covmap.coverage.data[idx] += 1;
        }

        if (isForward && !isLetterspace)
        {
            int idx = (haplome - 1) * covmap.coverage.size_x +
                      seqrange.data[0] - 2;
            VA_ASSERT(idx < covmap.coverage.size_x * covmap.coverage.size_y);
            covmap.coverage.data[idx] += 1;
        }
 		//fprintf(stderr,"9:%p\n",logerr.data);
        free(logerr.data);
        //fprintf(stderr,"10:%p\n",logcor.data);
        free(logcor.data);
        //fprintf(stderr,"11:%p\n",err.data);
        free(err.data);
        if (ferr.data != NULL)
        	//fprintf(stderr,"12:%p\n",ferr.data);
            free(ferr.data);
        if (seqrange.data != NULL)
        	//fprintf(stderr,"13:%p\n",seqrange.data);
            free(seqrange.data);
        usedreads++;
    }

    /* print time information */
    rc = clock_gettime(CLOCK_REALTIME, &te);
    VA_ASSERT(rc == 0);

    tdiff = va_TimeSub(&te, &ts);
    fprintf(stderr, "\tcovmaps built in: %ld.%03ld secs, "
        "Used reads in the covmaps: %d/%d, Skipped for 0 index:%d\n",
        tdiff.tv_sec, tdiff.tv_nsec/NS_IN_MS,
        usedreads, cdata.reads.size, zeroindexskippedreads);

    return (covmap);
} 


