/* 
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <libgen.h>
#include <stdio.h>
#include <errno.h>
#include <getopt.h>
#include <sys/stat.h>
#include <string.h>
#include "va_log.h"
#include "va_assert.h"
#include "varid.h"

#define VARID_VERSION  "0.99"

#define USAGE_TEXT "Usage: varid [ {-h | --help} ] [ {-v | --version} ]\n"  \
   "      [ --alignments ] [ --ref-file ] [ --donor-name ]\n"          \
   "      [ --output ] [ --temp-dir ] [ -s for sam format alginments ]\n"     

#define HELP_TEXT USAGE_TEXT "\n"                          \
    "\t\t-h, --help          \tthis help output\n"         \
    "\t\t-v, --version       \toutput version string\n"    \
    "\t\t-s, --sam           \talignments is in sam format\n"\
    "\t\t    --alignments    \talignments file name\n"               \
    "\t\t    --ref-file      \treference file name\n"           \
    "\t\t    --donor-name      \tdonor name\n"           \
    "\t\t    --output\twrite pretty file\n"        \
    "\t\t    --temp-dir       \ttemporary folder\n"      

#define VERSION_TEXT "varid " VARID_VERSION "\n"    \
    "Contact <varid at cs dot toronto dot edu>.\n\n" \
    "Copyright (C) 2010\n"                                    \
    "This program comes with ABSOLUTELY NO WARRANTY;\n" \
    "This is free software, and you are welcome to redistribute it\n" \
    "under certain conditions.\n\n Author: Adrian V. Dalca\n"

static char version[] = VERSION_TEXT;

static char * pname = NULL;

static void parse_args(
    int    argc,
    char * argv[]);

static void print_usage(
    int          exit_code,
    const char * s);

static void validate_args(void);

static int useSam = 0;

char * alignmentsFile = NULL;
char * refFile = NULL;
char * donName = NULL;
char * wrPrettyFile = NULL;
char * tempPath = NULL;

/*Probably there is a better place for these, but I don't know it */
double*  cs_error_cnt;
double*  cs_error_dpts;
unsigned int cs_error_size;
double*  ls_error_cnt;
double*  ls_error_dpts;
unsigned int ls_error_size;

/* program name */
static char * progName = NULL;

/* Called upon application exit */
static void varid_exit(void);

int main(
    int    argc, 
    char * argv[])
{
    VA_VecDouble thr_vec_p = va_allocVecDouble(1);

    progName = basename(argv[0]);

    parse_args(argc, argv);
  
    validate_args();

    /* Open the syslog */
    LOG_OPEN(progName);

    /* Register a function to be called when the application exit()s */
    (void)atexit(varid_exit);

    thr_vec_p.data[0] = 0.98;
    varid_single(alignmentsFile, refFile, donName, thr_vec_p, tempPath, useSam);

    exit (EXIT_SUCCESS);
}

static void parse_args(
    int    argc,
    char * argv[])
{
    char short_op[] = "hvs";
    struct option long_op[] =
    {
        {"help", 0, 0, 0},
        {"version", 0, 0, 0},
        {"alignments", 1, 0, 0},
        {"ref-file", 1, 0, 0},
        {"donor-name", 1, 0, 0},
        {"output", 1, 0, 0},
        {"temp-dir", 1, 0, 0},
        {"sam", 0, 0, 0},
        {0, 0, 0, 0}
    };
    int op_id;
    char c;

    c = getopt_long(argc, argv, short_op, long_op, &op_id);
    while (c != EOF)
    {
        switch (c)
        {
        case 0:
            /* long op format */
            switch (op_id)
            {
            case 0: /* help */
                printf ("%s\n", HELP_TEXT);
                exit (EXIT_SUCCESS);
                break;

            case 1: /* version */
                printf ("%s\n", version);
                exit (EXIT_SUCCESS);
                break;

            case 2: /* reads file */
                if (optarg == NULL)
                    print_usage(1, "Missing `alignments file name` after `--alignments`");
                else
                    alignmentsFile = strdup(optarg);
                break;

            case 3: /* reference file */
                if (optarg == NULL)
                    print_usage(1, "Missing `reference file name` after "
                                   "`--ref-file`");
                else
                    refFile = strdup(optarg);
                break;
 
            case 4: /* donor name */
                if (optarg == NULL)
                    print_usage(1, "Missing `donor genome name` after "
                                   "`--donor-name`");
                else
                    donName = strdup(optarg);
                break;

            case 5: /* write pretty file */
                if (optarg == NULL)
                    print_usage(1, "Missing `write pretty file` after "
                                   "`--output`");
                else
                    wrPrettyFile = strdup(optarg);
                break;

            case 6: /* temp-dir */
                if (optarg == NULL)
                    print_usage(1, "Missing `temporary file path` after `--temp-dir`");
                else
                    tempPath = strdup(optarg);
                break;

            case 7: /* sam */
            	useSam = 1;
            	break;

            default: /* unknown option */
                break;
            }
            break;

        case 'h':
            printf ("%s\n", HELP_TEXT);
            exit (EXIT_SUCCESS);
            break;

        case 'v':
            printf ("%s\n", version);
            exit(EXIT_SUCCESS);
            break;

        case 's':
			useSam = 1;
			break;

        default:
            break;
        }
        c = getopt_long(argc, argv, short_op, long_op, &op_id);
    }
}

static void print_usage(
    int          exit_code,
    const char * s)
{
    FILE * f;
    if (exit_code != 0)
        f = stderr;
    else
        f = stdout;

    if (s != NULL)
        fprintf(f, "%s: %s\n\n", progName, s);

    fprintf(f, "%s\n%s: %s\n", version, progName, USAGE_TEXT);
    exit (exit_code);
}

static void varid_exit(
    void)
{
    LOG_INFO1("%s exiting...", progName);    

    /* close syslog */
    LOG_CLOSE(progName);
}

static void validate_args(void)
{
    char buf[BUFSIZ];
    struct stat fstat;

    if (alignmentsFile == NULL)
    {
        print_usage(1, "Missing `alignments file`");
    }
    else if (stat(alignmentsFile, &fstat) == -1)
    {
        sprintf(buf, "Missing `alignments file`: %s", alignmentsFile);
        print_usage(1, buf);
    }

    if (refFile == NULL)
    {
        print_usage(1, "Missing `reference file`");
    }
    else if (stat(refFile, &fstat) == -1)
    {
        sprintf(buf, "Missing `reference file`: %s", refFile);
        print_usage(1, buf);
    }

    if (donName == NULL) {
	donName = "DONOR_GENOME";
        //print_usage(1, "Missing `reference name`");
    }

    if (tempPath == NULL)
    {
        print_usage(1, "Missing `temporary path`");
    }
    else if (stat(tempPath, &fstat) == -1)
    {
        sprintf(buf, "Missing `temporary path`: %s", tempPath);
        print_usage(1, buf);
    }

}
