/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

#define CHECK_RETURN(_a)    \
    do                      \
    {                       \
        if ((_a) <= 0)      \
             return (0);    \
    } while (0);

static VA_Vector va_loadReads(
    char   * filename,
    char   * rdmat,
    size_t * readsSize);

static VA_VecI32 va_loadIns(
    char   * filename,
    char   * insmat,
    size_t * insSize);

static size_t va_saveReads(
    VA_Vector  reads,
    size_t     leftlen,
    char     * stream);

static size_t va_saveInsVec(
    VA_VecI32  insvec,
    size_t     leftLen,
    char     * stream);

VA_UINT32 tot_saved = 0;

void va_load(
    const char * filename,
    VA_Vector  * reads,
    VA_VecI32  * insvec)
{
    char line[BUFSIZ];
    char * rdmat = NULL;
    char * insmat = NULL;
    size_t nsize = (size_t)-1;
    int totalReadsSize;
    size_t readsSize = 0;
    int totalInsSize;
    size_t insSize = 0;
    VA_Vector rdCtrl;
    VA_VecI32 insCtrl;
    FILE       * fidmat;

    VA_ASSERT(filename != NULL);
    VA_ASSERT(reads != NULL);
    VA_ASSERT(insvec != NULL);

    fidmat = fopen(filename, "r");
    VA_ASSERT(fidmat != NULL);

    reads->size = 0;
    reads->data = NULL;
    insvec->size = 0;
    insvec->data = NULL;

    nsize = fread(line, 1, sizeof(VA_UINT32), fidmat);
    if (nsize != sizeof(VA_UINT32))
    {
       LOG_ERROR1("File mat <%s>: Invalid header", filename);
       exit(EXIT_FAILURE);
    }

    totalReadsSize = *((VA_UINT32 *)line);

    rdmat = (char *)malloc(totalReadsSize);
    VA_ASSERT(rdmat != NULL);

    nsize = fread(rdmat, 1, totalReadsSize, fidmat);
    if (nsize != totalReadsSize)
    {
       free(rdmat);
       LOG_ERROR1("File matc <%s>: reads size", filename);
       exit(EXIT_FAILURE);
    }

    VA_ASSERT(totalReadsSize > sizeof(VA_UINT32));

    rdCtrl = va_loadReads((char *)filename, rdmat, &readsSize);
    VA_ASSERT(totalReadsSize == readsSize); 
    reads->size = rdCtrl.size;
    reads->data = rdCtrl.data;
    
    nsize = fread(line, 1, sizeof(VA_UINT32), fidmat);
    if (nsize != sizeof(VA_UINT32))
    {
       LOG_ERROR1("File matc <%s>: Ins header", filename);
       exit(EXIT_FAILURE);
    }

    totalInsSize = *((VA_UINT32 *)line);

    if (totalInsSize > 0)
    {
        insmat = (char *)malloc(totalInsSize);
        VA_ASSERT(insmat != NULL);

        nsize = fread(insmat, 1, totalInsSize, fidmat);
        if (nsize != totalInsSize)
        {
           LOG_ERROR1("File matc <%s>: reads size", filename);
           exit(EXIT_FAILURE);
        }

        insCtrl = va_loadIns((char *)filename, insmat, &insSize);
        VA_ASSERT(totalInsSize == insSize); 
        insvec->size = insCtrl.size;
        insvec->data = insCtrl.data;

        free(insmat);
    }
    else
    {
        insvec->size = 0;
        insvec->data = NULL;
    }
    
    free(rdmat);
    fclose(fidmat);
}

static VA_VecI32 va_loadIns(
    char   * filename,
    char   * insmat,
    size_t * insSize)
{ 
    int i;
    VA_VecI32 ret;
    VA_UINT32 * q;
    char * p = insmat;

    VA_ASSERT(p != NULL);
    VA_ASSERT(filename != NULL);
    VA_ASSERT(insSize != NULL);

    /* read dimension */
    ret.size = *(VA_UINT32 *)p;
    ret.data = NULL;
    p += sizeof(VA_UINT32);

    if (ret.size > 0)
    {
        int len = *(VA_UINT16 *)p;

        VA_ASSERT(len == sizeof(VA_UINT32) * ret.size);
        p += sizeof(VA_UINT16);
        ret.data = (VA_UINT32 *)malloc(ret.size * sizeof(VA_UINT32));
        VA_ASSERT(ret.data != NULL);

        q = (VA_UINT32 *)ret.data;

        for (i = 0; i < ret.size; i++)
        {
            q[i] = *(VA_UINT32 *)p;
            p += sizeof(VA_UINT32);
        }
    }

    *insSize = (size_t)(p - insmat);
    return (ret);
}

static VA_Vector va_loadReads(
    char   * filename,
    char   * rdmat,
    size_t * readsSize)
{ 
    ReadsNode_t * rd = NULL;
    VA_UINT32  nreads = 0;
    int pos;
    int i;
    VA_Vector ret;

    char * p = rdmat;

    VA_ASSERT(rdmat != NULL);
    VA_ASSERT(filename != NULL);
    VA_ASSERT(readsSize != NULL);

    /* read second dimension */
    nreads = *(VA_UINT32 *)p;
    p += sizeof(VA_INT32);

    rd = (ReadsNode_t *)malloc(sizeof(ReadsNode_t) * nreads); 
    VA_ASSERT(rd != NULL);

    pos = (VA_UINT32)(p - rdmat);

    for (i = 0; i < nreads; i++)
    {
        if (*(VA_UINT16 *)p == 0)
        {
            rd[i].name = NULL;
        }
        else
        {
            rd[i].name = (VA_UINT8 *)malloc(sizeof(VA_UINT8) * (*(VA_UINT16 *)p) + 1); 
            VA_ASSERT(rd[i].name != NULL);
            memcpy(rd[i].name, p + sizeof(VA_UINT16), *(VA_UINT16 *)p);
            rd[i].name[(*(VA_UINT16 *)p)] = 0;
        }
        p += sizeof(VA_UINT16) + (*(VA_UINT16 *)p);
        
        rd[i].strand = (char)*(VA_UINT32 *)p;
        p += sizeof(VA_UINT32);

        rd[i].chrstart = *(VA_UINT32 *)p;
        p += sizeof(VA_UINT32);

        rd[i].chrend = *(VA_UINT32 *)p;
        p += sizeof(VA_UINT32);

        if (*(VA_UINT16 *)p == 0)
        {
            rd[i].letseq.data = NULL;
            rd[i].letseq.size = 0;
        }
        else
        {
            rd[i].letseq.size = *(VA_UINT16 *)p;
            rd[i].letseq.data = 
                 (VA_UINT8 *)malloc(sizeof(VA_UINT8) * rd[i].letseq.size); 
            VA_ASSERT(rd[i].letseq.data != NULL);
            memcpy(rd[i].letseq.data, p + sizeof(VA_UINT16), 
                sizeof(VA_UINT8) * rd[i].letseq.size);
        }
        p += sizeof(VA_UINT16) + sizeof(VA_UINT8) * rd[i].letseq.size;

        if (*(VA_UINT16 *)p == 0)
        {
            rd[i].colseq.data = NULL;
            rd[i].colseq.size = 0;
        }
        else
        {
            rd[i].colseq.size = *(VA_UINT16 *)p;
            rd[i].colseq.data = 
                 (VA_UINT8 *)malloc(sizeof(VA_UINT8) * rd[i].colseq.size); 
            VA_ASSERT(rd[i].colseq.data != NULL);
            memcpy(rd[i].colseq.data, p + sizeof(VA_UINT16), 
                sizeof(VA_UINT8) * rd[i].colseq.size);
        }
        p += sizeof(VA_UINT16) + sizeof(VA_UINT8) * rd[i].colseq.size;

        if (*(VA_UINT16 *)p == 0)
        {
            rd[i].ins.data = NULL;
            rd[i].ins.size = 0;
        }
        else
        {
            rd[i].ins.size = *(VA_UINT16 *)p;
            rd[i].ins.data = 
                 (VA_UINT32 *)malloc(sizeof(VA_UINT32) * rd[i].ins.size); 
            VA_ASSERT(rd[i].ins.data != NULL);
            memcpy(rd[i].ins.data, p + sizeof(VA_UINT16), 
                sizeof(VA_UINT32) * rd[i].ins.size);
        }
        p += sizeof(VA_UINT16) + sizeof(VA_UINT32) * rd[i].ins.size;

        rd[i].seqtype = *(VA_UINT32 *)p;
        p += sizeof(VA_UINT32);

        rd[i].extras.haplome = *(VA_UINT32 *)p;
        p += sizeof(VA_UINT32);

        rd[i].extras.score = *(VA_UINT32 *)p;
        p += sizeof(VA_UINT32);

        memcpy(&rd[i].extras.nodds, p, sizeof(VA_DOUBLE));
        p += sizeof(VA_DOUBLE);

        memcpy(&rd[i].extras.pgenome, p, sizeof(VA_DOUBLE));
        p += sizeof(VA_DOUBLE);

        memcpy(&rd[i].extras.pchance, p, sizeof(VA_DOUBLE));
        p += sizeof(VA_DOUBLE);
    }

    ret.size = nreads;
    ret.data = rd;

    *readsSize = (int)(p - rdmat);
    return (ret);
}

void va_loadMatDouble(
    char         * filename,
    VA_MatDouble * p)
{
    FILE * fidmat;
    size_t sz;
    size_t nsize;
    VA_UINT32 px = 0;
    VA_UINT32 py = 0;

    fidmat = fopen(filename, "r");
    VA_ASSERT(fidmat != NULL);

    nsize = fread(&px, 1, sizeof(VA_UINT32), fidmat);
    if (nsize != sizeof(VA_UINT32))
    {
       LOG_ERROR1("File <%s>: invalid control data", filename);
       exit(EXIT_FAILURE);
    }

    nsize = fread(&py, 1, sizeof(VA_UINT32), fidmat);
    if (nsize != sizeof(VA_UINT32))
    {
       LOG_ERROR1("File <%s>: invalid control data", filename);
       exit(EXIT_FAILURE);
    }

    sz = px * py * sizeof(VA_DOUBLE);
    *p = va_allocMatDouble(px, py);
    nsize = fread(p->data, 1, sz, fidmat);
    if (nsize != sz)
    {
       LOG_ERROR1("File <%s>: invalid data size", filename);
       exit(EXIT_FAILURE);
    }
    
    fclose(fidmat);
}

void va_saveMatDouble(
    char        * filename,
    VA_MatDouble  p)
{
    char * buf2mat = NULL;
    VA_INT32 buf2matLen = 0;
    size_t pos = 0;
    FILE        * fidmat;
    int i;
    int j;

    buf2matLen = p.size_x * p.size_y * sizeof(VA_DOUBLE) + 2 * sizeof(VA_UINT32);
 
    fidmat = fopen(filename, "a");
    VA_ASSERT(fidmat != NULL);

    buf2mat = (char *)malloc(buf2matLen); 

    *((VA_UINT32 *)&buf2mat[pos]) = p.size_x; 
    pos += sizeof(VA_UINT32);

    *((VA_UINT32 *)&buf2mat[pos]) = p.size_y; 
    pos += sizeof(VA_UINT32);

    for (i = 0; i < p.size_y; i++)
    {
        for (j = 0; j < p.size_x; j++)
        {
           *((VA_DOUBLE *)&buf2mat[pos]) = p.data[i * p.size_x + j];
           pos += sizeof(VA_DOUBLE);
        }
    }

    fwrite(buf2mat, 1, buf2matLen, fidmat);
    free(buf2mat);
    fclose(fidmat);
}
      
void va_save(
    const char * filename,
    VA_Vector   reads,
    VA_VecI32   insvec)
{
    char * buf2mat = NULL;
    VA_INT32 buf2matTotalLen = 0;
    VA_INT32 buf2matCurrLen = 0;
    size_t pos = 0;
    FILE        * fidmat;

    buf2matTotalLen = reads.size * sizeof(ReadsNode_t) * 50;
 
    fidmat = fopen(filename, "a");
    VA_ASSERT(fidmat != NULL);

    while (pos == 0)
    {
        buf2mat = (char *)malloc(buf2matTotalLen);
        VA_ASSERT(buf2mat != NULL); 
        buf2matCurrLen = 0;

        pos = va_saveReads(reads, buf2matTotalLen, buf2mat);
        if (pos <= 0)
        {
            free(buf2mat);
            buf2matTotalLen *= 10;
            continue;
        }
      
        buf2matCurrLen += pos;
        pos = va_saveInsVec(insvec, buf2matTotalLen - buf2matCurrLen, 
                  buf2mat + buf2matCurrLen);
        if (pos <= 0)
        {
            free(buf2mat);
            buf2matTotalLen *= 10;
        }
        buf2matCurrLen += pos;
    }

    fwrite(buf2mat, 1, buf2matCurrLen, fidmat);
    free(buf2mat);
    fclose(fidmat);
}

static size_t va_saveReads(
    VA_Vector     reads,
    size_t        leftLen,
    char        * stream)
{
    VA_UINT32 pos = 0;
    VA_UINT32 curr_read = 0;
    size_t pos2 = -1;
    ReadsNode_t * rdata = (ReadsNode_t *)reads.data;

    VA_ASSERT(reads.data != NULL);
    VA_ASSERT(reads.size > 0);
    VA_ASSERT(stream != NULL);

    *((VA_UINT32 *)&stream[pos]) = 0; /* total matrix len */
    pos += sizeof(VA_UINT32);

    *((VA_UINT32 *)&stream[pos]) = reads.size; /* reads second dimension */
    pos += sizeof(VA_UINT32);

    while ((pos > 0) && (curr_read < reads.size))
    {
        ReadsNode_t * rd = &rdata[curr_read++];   
        
        pos2 = pos + sizeof(VA_UINT16) + strlen(rd->name);
        CHECK_RETURN(leftLen - pos2)
        *((VA_UINT16 *)&stream[pos]) = strlen(rd->name);
        if (rd->name[0] != 0)
            memcpy(&stream[pos + sizeof(VA_UINT16)], rd->name, 
                strlen(rd->name));
        pos = pos2; 
        
        pos2 = pos + sizeof(VA_UINT32);
        CHECK_RETURN(leftLen - pos2)
        memcpy(&stream[pos], &rd->strand, sizeof(rd->strand));
        pos = pos2;
       
        pos2 = pos + sizeof(VA_UINT32);
        CHECK_RETURN(leftLen - pos2)
        memcpy(&stream[pos], &rd->chrstart, sizeof(rd->chrstart));
        pos = pos2;
       
        pos2 = pos + sizeof(VA_UINT32);
        CHECK_RETURN(leftLen - pos2)
        memcpy(&stream[pos], &rd->chrend, sizeof(rd->chrend));
        pos = pos2;
       
        pos2 = pos + sizeof(VA_UINT16) + rd->letseq.size * sizeof(VA_UINT8);
        CHECK_RETURN(leftLen - pos2)
        if (rd->letseq.size > 0)
            memcpy(&stream[pos + sizeof(VA_UINT16)], rd->letseq.data, 
                rd->letseq.size * sizeof(VA_UINT8));
        *((VA_UINT16 *)&stream[pos]) = rd->letseq.size;
        pos = pos2;
       
        pos2 = pos + sizeof(VA_UINT16) + rd->colseq.size * sizeof(VA_UINT8);
        CHECK_RETURN(leftLen - pos2)
        if (rd->colseq.size > 0)
            memcpy(&stream[pos + sizeof(VA_UINT16)], rd->colseq.data, 
                rd->colseq.size * sizeof(VA_UINT8));
        *((VA_UINT16 *)&stream[pos]) = rd->colseq.size;
        pos = pos2;
       
        pos2 = pos + sizeof(VA_UINT16) + rd->ins.size * sizeof(VA_UINT32);
        CHECK_RETURN(leftLen - pos2)
        if (rd->ins.size > 0)
            memcpy(&stream[pos + sizeof(VA_UINT16)], rd->ins.data, 
                rd->ins.size * sizeof(VA_UINT32));
        *((VA_UINT16 *)&stream[pos]) = rd->ins.size;
        pos = pos2;
       
        pos2 = pos + sizeof(VA_UINT32);
        CHECK_RETURN(leftLen - pos2)
        memcpy(&stream[pos], &rd->seqtype, sizeof(rd->seqtype));
        pos = pos2;

        pos2 = pos + sizeof(VA_UINT32);
        CHECK_RETURN(leftLen - pos2)
        memcpy(&stream[pos], &rd->extras.haplome, sizeof(rd->extras.haplome));
        pos = pos2;

        pos2 = pos + sizeof(VA_UINT32);
        CHECK_RETURN(leftLen - pos2)
        memcpy(&stream[pos], &rd->extras.score, sizeof(rd->extras.score));
        pos = pos2;
       
        pos2 = pos + sizeof(VA_DOUBLE);
        CHECK_RETURN(leftLen - pos2)
        memcpy(&stream[pos], &rd->extras.nodds, sizeof(rd->extras.nodds));
        pos = pos2;
       
        pos2 = pos + sizeof(VA_DOUBLE);
        CHECK_RETURN(leftLen - pos2)
        memcpy(&stream[pos], &rd->extras.pgenome, sizeof(rd->extras.pgenome));
        pos = pos2;
       
        pos2 = pos + sizeof(VA_DOUBLE);
        CHECK_RETURN(leftLen - pos2)
        memcpy(&stream[pos], &rd->extras.pchance, sizeof(rd->extras.pchance));
        pos = pos2;
    }
    
    if (pos > 0)
    {
        *(VA_UINT32 *)stream = pos - sizeof(VA_UINT32);
        tot_saved += pos;
    }

    return (pos);
}

static size_t va_saveInsVec(
    VA_VecI32    insvec,
    size_t       leftLen,
    char       * stream)
{
    int pos = 0;
    int pos2 = 0;

    VA_ASSERT((insvec.size == 0) || 
       ((insvec.size > 0) && (insvec.data != NULL)));
    VA_ASSERT(stream != NULL);

    *((VA_UINT32 *)&stream[pos]) = 0; /* total matrix len */
    pos += sizeof(VA_UINT32);

    *((VA_UINT32 *)&stream[pos]) = insvec.size; /* reads second dimension */
    pos += sizeof(VA_UINT32);

    if (insvec.size > 0)
    {
        pos2 = pos + sizeof(VA_UINT16) + insvec.size * sizeof(VA_UINT32);
        CHECK_RETURN(leftLen - pos2)
        memcpy(&stream[pos + sizeof(VA_UINT16)], insvec.data,
            insvec.size * sizeof(VA_UINT32));
        *((VA_UINT16 *)&stream[pos]) = insvec.size * sizeof(VA_UINT32);
        pos = pos2;
    }

    if (pos > 0)
    {
        *(VA_UINT32 *)stream = pos - sizeof(VA_UINT32);
        tot_saved += pos;
    }

    return (pos);
}


