/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

#define FASTA_DEFAULT_NUM_RECS (10)

static int cmpSortI32(
    const void * rec1,
    const void * rec2);

VA_VecI32 sortVecI32;

char * va_basename(
   const char * file)
{
   char * fname = basename((char *)file);
   char * ret;
   char * p;
   int nsz;

   p = strrchr(fname, '.');
   if (p != NULL)
       nsz = p - fname;
   else
       nsz = strlen(fname);
   ret = (char *)malloc(nsz + 1);

   memcpy(ret, fname, nsz);
   ret[nsz] = 0;

   return (ret);
}

int va_readfasta(
    const char * reffile,
    VA_Vector  * vfasta)
{
    VA_Vector fasta;
    FILE * fid;
    char * p;
    char * q;
    char * t;
    int hdrlen = 0;
    int seqlen = 0;
    VA_UINT32 pos = 0;
    VA_UINT32 n;
    VA_UINT32 fastaIdx = 0;
    Fasta_t * f;
    int state;
    int ncrlf;
    size_t fileLength;
    int okSoFar = TRUE;

    VA_ASSERT(reffile != NULL);
    fid = fopen(reffile, "r");
    VA_ASSERT(fid != NULL);
    fseek(fid, 0, SEEK_END);
    fileLength = ftell(fid) + 1;
    fseek(fid, 0, SEEK_SET);

    p = malloc(fileLength);
    VA_ASSERT(p != NULL);

    fasta.size = FASTA_DEFAULT_NUM_RECS;
    fasta.data = malloc(sizeof(Fasta_t) * fasta.size);
    VA_ASSERT(fasta.data != NULL);
    f = (Fasta_t *)fasta.data;

    n = fread(p, fileLength, 1, fid);
    VA_ASSERT(n == 0);

    p[fileLength - 1] = 0;

    state = 0;
    while ((okSoFar == TRUE) && (pos < fileLength - 1))
    {
        switch (state)
        {
            case 0:
                if (p[pos] != '>')
                {
                    LOG_ERROR1("fasta file <%s>: missing header", reffile);
                    okSoFar = FALSE;
                }
                else
                {
                    hdrlen = 0;
                    if (fastaIdx >= fasta.size)
                    {
                        VA_Vector tmpfasta;

                        tmpfasta.size = fasta.size * 2;
                        tmpfasta.data = malloc(sizeof(Fasta_t) * tmpfasta.size);
                        VA_ASSERT(tmpfasta.data != NULL);

                        memcpy(tmpfasta.data, fasta.data,
                            sizeof(Fasta_t) * fasta.size);
                        free(fasta.data);
                        memcpy(&fasta, &tmpfasta, sizeof(VA_Vector));
                        f = (Fasta_t *)fasta.data;
                    }

                    pos++;
                    q = &p[pos];

                    while ((p[pos] != 0)
                        && (p[pos] != '\n')
                        && (p[pos] != '\r'))
                    {
                        pos++;
                        hdrlen++;
                    }

                    if (p[pos] == 0)
                    {
                        LOG_ERROR1("fasta file <%s>: header without sequence",
                            reffile);
                        okSoFar = FALSE;
                        break;
                    }

                    pos++;
                    while ((p[pos] == '\n') || (p[pos] == '\r'))
                        pos++;

                    if (p[pos] == 0)
                    {
                        LOG_ERROR1("fasta file <%s>: header without sequence",
                            reffile);
                        okSoFar = FALSE;
                        break;
                    }
                    f[fastaIdx].hdr = (char *)malloc(hdrlen + 1);
                    VA_ASSERT(f[fastaIdx].hdr != NULL);

                    memcpy(f[fastaIdx].hdr, q, hdrlen);
                    f[fastaIdx].hdr[hdrlen] = 0;

                    state = 1;
                }

                break;
            case 1:
                seqlen = 0;
                ncrlf = 0;
                q = &p[pos];

                while ((p[pos] != 0) && (p[pos] != '>'))
                {
                    if ((p[pos] == '\n') || (p[pos] == '\r'))
                        ncrlf++;
                    pos++;
                    seqlen++;
                }

                seqlen -= ncrlf;
                f[fastaIdx].seq = va_allocVecU8(seqlen + 1);
                t = f[fastaIdx].seq.data;

                while (seqlen > 0)
                {
                    if ((*q != '\n') && (*q != '\r'))
                    {
                       *t++ = *q;
                       seqlen--;
                    }
                    q++;
                }

                *t = 0;
                fastaIdx++;
                state = 0;
                break;
        }
    }

    vfasta->data = malloc(sizeof(Fasta_t) * fastaIdx);
    VA_ASSERT(vfasta->data!= NULL);
    vfasta->size = fastaIdx;
    
    memcpy(vfasta->data, fasta.data, sizeof(Fasta_t) * vfasta->size);
    fclose(fid);

    return (okSoFar == TRUE) ? 0 : 1;
}

void fliplrU8(
    VA_VecU8 toFlip)
{
    int i;

    VA_ASSERT((toFlip.data != NULL) && (toFlip.size > 0));

    for (i = 0; i < toFlip.size / 2; i++)
    {
        VA_UINT8 ch = toFlip.data[i];
        toFlip.data[i] = toFlip.data[toFlip.size - 1 - i];
        toFlip.data[toFlip.size - 1 - i] = ch;
    }
}

void fliplrDouble(
    VA_VecDouble toFlip)
{
    int i;

    VA_ASSERT((toFlip.data != NULL) && (toFlip.size > 0));

    for (i = 0; i < toFlip.size / 2; i++)
    {
        VA_DOUBLE ch = toFlip.data[i];
        toFlip.data[i] = toFlip.data[toFlip.size - 1 - i];
        toFlip.data[toFlip.size - 1 - i] = ch;
    }
}

VA_VecU8 va_reallocVecU8(
    VA_VecU8 vec,
    size_t   n)
{
    VA_VecU8 ret;

    VA_ASSERT((vec.data != NULL) && (vec.size > 0));
    VA_ASSERT(n > 0);

    ret.size = n;
    ret.data = (VA_UINT8 *)malloc(sizeof(VA_UINT8) * ret.size);
    VA_ASSERT(ret.data != NULL);

    memcpy(ret.data, vec.data, (n > vec.size) ? vec.size : n);
    free(vec.data);

    return (ret);
}

VA_VecU32 va_reallocVecU32(
    VA_VecU32 vec,
    size_t      n)
{
    VA_VecU32 ret;

    VA_ASSERT((vec.data != NULL) && (vec.size > 0));
    VA_ASSERT(n > 0);

    ret.size = n;
    ret.data = (VA_UINT32 *)malloc(sizeof(VA_UINT32) * ret.size);
    VA_ASSERT(ret.data != NULL);

    memcpy(ret.data, vec.data, 
        (n > vec.size) ? vec.size * sizeof(VA_UINT32) : n * sizeof(VA_UINT32));
    free(vec.data);

    return (ret);
}

VA_VecI32 va_reallocVecI32(
    VA_VecI32 vec,
    size_t      n)
{
    VA_VecI32 ret;

    VA_ASSERT((vec.data != NULL) && (vec.size > 0));
    VA_ASSERT(n > 0);

    ret.size = n;
    ret.data = (VA_INT32 *)malloc(sizeof(VA_INT32) * ret.size);
    VA_ASSERT(ret.data != NULL);

    memcpy(ret.data, vec.data, 
        (n > vec.size) ? vec.size * sizeof(VA_INT32) : n * sizeof(VA_INT32));
    free(vec.data);

    return (ret);
}

VA_VecU8 va_allocVecU8(
    size_t  size)
{
    VA_VecU8 ret;

    ret.size = size;
    ret.data = (VA_UINT8 *)malloc(sizeof(VA_UINT8) * size);
    VA_ASSERT(ret.data != NULL);

    return (ret);
}

VA_VecI32 va_allocVecI32(
    size_t  size)
{
    VA_VecI32 ret;

    ret.size = size;
    ret.data = (VA_INT32 *)malloc(sizeof(VA_INT32) * size);
    VA_ASSERT(ret.data != NULL);

    return (ret);
}

VA_VecU32 va_allocVecU32(
    size_t  size)
{
    VA_VecU32 ret;

    ret.size = size;
    ret.data = (VA_UINT32 *)malloc(sizeof(VA_UINT32) * size);
    VA_ASSERT(ret.data != NULL);

    return (ret);
}

VA_VecDouble va_allocVecDouble(
    size_t  size)
{
    VA_VecDouble ret;

    ret.size = size;
    ret.data = (VA_DOUBLE *)malloc(sizeof(VA_DOUBLE) * size);
    VA_ASSERT(ret.data != NULL);

    return (ret);
}

VA_MatI32 va_allocMatI32(
    size_t  size_x,
    size_t  size_y)
{
    VA_MatI32 ret;

    ret.size_x = size_x;
    ret.size_y = size_y;
    ret.data = (VA_INT32 *)malloc(sizeof(VA_INT32) * size_x * size_y);
    VA_ASSERT(ret.data != NULL);

    return (ret);
}

VA_MatU32 va_allocMatU32(
    size_t  size_x,
    size_t  size_y)
{
    VA_MatU32 ret;

    ret.size_x = size_x;
    ret.size_y = size_y;
    ret.data = (VA_UINT32 *)malloc(sizeof(VA_UINT32) * size_x * size_y);
    VA_ASSERT(ret.data != NULL);

    return (ret);
}

VA_MatU8 va_allocMatU8(
    size_t  size_x,
    size_t  size_y)
{
    VA_MatU8 ret;

    ret.size_x = size_x;
    ret.size_y = size_y;
    ret.data = (VA_UINT8 *)malloc(sizeof(VA_UINT8) * size_x * size_y);
    VA_ASSERT(ret.data != NULL);

    return (ret);
}

VA_MatDouble va_allocMatDouble(
    size_t  size_x,
    size_t  size_y)
{
    VA_MatDouble ret;

    ret.size_x = size_x;
    ret.size_y = size_y;
    ret.data = (VA_DOUBLE *)malloc(sizeof(VA_DOUBLE) * size_x * size_y);
    VA_ASSERT(ret.data != NULL);

    return (ret);
}

VA_VecI32 va_unique(
    VA_VecI32 ref)
{
    VA_VecI32 ret;
    int j;

    VA_ASSERT((ref.data != NULL) && (ref.size > 0));
    ret.data = (VA_INT32 *)malloc(ref.size * sizeof(VA_INT32));
    VA_ASSERT(ret.data != NULL);
    ret.size = 0;

    for (j = 0; j < ref.size; j++)
    {
        int k;
        int isPresent = 0;

        for (k = 0; k < ret.size; k++)
        {
            if (ret.data[k] == ref.data[j])
            {
                isPresent = 1;
                break;
            }
        }

        if (isPresent == 0)
            ret.data[ret.size++] = ref.data[j];
    }

    return (ret);
}

VA_VecI32 va_sortVecI32(
    VA_VecI32 vec)
{
    VA_VecI32 idx = va_allocVecI32(vec.size);
    VA_VecI32 ret = va_allocVecI32(vec.size);
    int i;

    for (i = 0; i < idx.size; i++)
        idx.data[i] = i;
    sortVecI32 = vec;

    qsort(idx.data, idx.size, sizeof(VA_INT32), cmpSortI32);

    for (i = 0; i < idx.size; i++)
        ret.data[i] = vec.data[idx.data[i]];

    free(idx.data);
    free(vec.data);

    return (ret);
}

static int cmpSortI32(
    const void * rec1,
    const void * rec2)
{
    VA_INT32 p1;
    VA_INT32 p2;

    VA_ASSERT((rec1 != NULL) && (rec2 != NULL));

    p1 = *(VA_INT32 *)rec1;
    p2 = *(VA_INT32 *)rec2;

    return (sortVecI32.data[p1] - sortVecI32.data[p2]);
}

