/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

extern char * summFile;
extern int singleGroup;
extern char * wrPrettyFile;

static void initThresholdVec(
    int            doEval,
    VA_VecDouble * thr_snp,
    VA_VecDouble * thr_p);

static char ** initGroupFiles(
    const char * readsfile,
    const char * path,
    int          nfiles);

static VA_Vector getFastaReference(
    const char  * reffile);

static int cmpSortI32(
   const void * rec1,
   const void * rec2);

char ** filenames = NULL;

/*******************************************************************************
* VARID_SINGLE prepare the data and run the VARiD algorithm.
*   varid_single(readsfile, reffile, don_name, thr_vec_p, path) will run
*   the varid algorithm after preparing the data. If the reference file
*   (mfa) will have more than one reference, the readsfile will be split
*   based on the references they are aligned to, in the folder <path>, and
*   the reads will be loaded piecewise based on the reference that is being
*   processed. The result will be outputed in prettybase format.
*
*   parameters:
*       readfile - path to the file of reads, which has to be in prettyprint
*           format for now...
*       reffile - path to the mfa reference file.
*       don_name - the name of the donor genome, to be included in the output
*       thr_vec_p - the probability threshold for p
*       path - the path for storing the remporary split of reads.
*
*   Author: Adrian V. Dalca
*   (adalca at mit dot edu)
*******************************************************************************/
void varid_single(
    const char   * readsfile, 
    const char   * reffile, 
    const char   * don_name, 
    VA_VecDouble   thr_vec_p,
    const char   * path,
    int			   useSam)
{
    FILE * fid;
    SnpGroups_t snpgroup;
    VA_UINT32 i;
    VA_UINT32 j;
    VA_UINT32 gnr;
    VA_INT32 splitreads;
    struct stat retstat;
    struct timespec mastertic;
    struct timespec readtic;
    struct timespec te;
    struct timespec tdiff;
    int rc;
    VA_UINT8 method;
    Fasta_t * fasta;
    VA_Vector fastaVec;
    Maps_t    maps;
    VA_Stats smallest_stats;
    VA_Stats stats;
    SnpGroup_t * p;
    VA_INT32 * fid_lines = NULL;
    VA_INT32 * fid_reads = NULL;
    VA_VecDouble   thr_vec_snp;

    VA_ASSERT(readsfile != NULL);
    VA_ASSERT(reffile != NULL);
    VA_ASSERT(don_name != NULL);
    VA_ASSERT(path != NULL);

    /* preparations */

    /* give some information to the user */
    fprintf(stderr, "\n\t - alignments: %s \n\t - ref-file:   %s\n\n"
        "Detected: \n", readsfile, reffile);

    /* prepare master clock */
    rc = clock_gettime(CLOCK_REALTIME, &mastertic);
    VA_ASSERT(rc == 0);

    /* threshold vectors */
    thr_vec_snp = va_allocVecDouble(1);
    thr_vec_snp.data[0] = 0.0;
       
    /* get the reference and shift snpgroups */
    fastaVec = getFastaReference(reffile);
    VA_ASSERT(fastaVec.size > 0);

    /* get the reads */
    rc = clock_gettime(CLOCK_REALTIME, &readtic);
    VA_ASSERT(rc == 0);
 
    fprintf(stderr, "Splitting reads\n");

    filenames = initGroupFiles(readsfile, path, fastaVec.size);

    snpgroup.size = fastaVec.size;
    snpgroup.data = (SnpGroup_t *)malloc(snpgroup.size * sizeof(SnpGroup_t));
    VA_ASSERT(snpgroup.data != NULL);
    memset(snpgroup.data, 0, sizeof(SnpGroup_t));
    p = (SnpGroup_t *)snpgroup.data;

    fasta = (Fasta_t *)fastaVec.data;

    for (i = 0; i < fastaVec.size; i++)
    {
        p[i].range = va_allocVecI32(2);
        p[i].range.data[0] = 1;
        p[i].range.data[1] = fasta[i].seq.size;
        p[i].refstart = 1;
        p[i].chr = va_allocVecU8(strlen(fasta[i].hdr));
        memcpy(p[i].chr.data, fasta[i].hdr, p[i].chr.size);
    }
            
    split_reads(readsfile, filenames, &snpgroup, 1, &fid_lines, &fid_reads, useSam, fastaVec);

    rc = clock_gettime(CLOCK_REALTIME, &te);
    VA_ASSERT(rc == 0);

    tdiff = va_TimeSub(&te, &readtic);

    /* done read prep */
    fprintf(stderr, "\t Done: in: %ld.%03ld secs\n",
        tdiff.tv_sec, tdiff.tv_nsec / NS_IN_MS);

    /* pre-process data and run VARiD on each SNP group */

    /* run processing and VARiD for every snp group */
    for (gnr = 0; gnr < snpgroup.size; gnr++)
    {
        if (filenames[gnr] != NULL)
        {
            VA_Vector reads;
            VA_VecI32 insvec;
            VA_INT32 startpos;
            VA_INT32 endpos;
            VA_INT32 * readstarts;
            VA_VecI32 f;
            VA_VecU8 g_ref;
            VA_Vector g_reads;
            VA_VecI32 g_readstarts;
            VA_VecI32 g_ins_vec; 
            VA_VecI32 t_ins_vec; 
            VA_VecI32 ins_vec_add; 
            CData_t cdata;
            ReadsNode_t * rd;
            int latest_pos;
            VA_MatDouble p;
            VA_MatDouble tot;

            fprintf(stderr, "------------------------------\n");
            fprintf(stderr, "Processing group number %d\n", gnr + 1);

            rc = clock_gettime(CLOCK_REALTIME, &readtic);
            VA_ASSERT(rc == 0);

            fprintf(stderr, "Loading reads ...\n");
            get_reads(filenames[gnr], fid_reads[gnr], fid_lines[gnr], 
                      &g_reads, &g_ins_vec);

            rc = clock_gettime(CLOCK_REALTIME, &te);
            VA_ASSERT(rc == 0);

            tdiff = va_TimeSub(&te, &readtic);
            fprintf(stderr, "\t Done in %ld.%03ld secs\n", 
                tdiff.tv_sec, tdiff.tv_nsec / NS_IN_MS);

            /* form the readstarts vector */
            g_readstarts = va_allocVecI32(g_reads.size);

            rd  = (ReadsNode_t *)g_reads.data;
            for (j = 0; j < g_reads.size; j++)
                g_readstarts.data[j] = rd[j].chrstart;  
             
            /* make sure there are reads in this snp group */
            if (g_reads.size == 0)
            {
                fprintf(stderr, "Group %d has no reads\n", gnr + 1);
                continue;
            }
      
            /* get the reads-generated insertion vector */
            if (g_ins_vec.size == 0)
            {
                g_ins_vec = va_allocVecI32(1);
                g_ins_vec.data[0] = 1;
            }
            
            /* prepare the insertion vector */
            if (g_ins_vec.size > 0)
                qsort(g_ins_vec.data, g_ins_vec.size, sizeof(VA_INT32), cmpSortI32);
                
            for (i = 1; i <= g_ins_vec.size; i++)
                g_ins_vec.data[i - 1] += i;

            if (g_ins_vec.size > 0)
            {
                VA_VecI32 su = va_unique(g_ins_vec);
                VA_ASSERT(su.size == g_ins_vec.size);
            } 
             
            /* extract needed reference and insert */
            {
               VA_VecU8 new_ref;

               g_ref = va_allocVecU8(fasta[gnr].ref.size);
               for (i = 0; i < g_ref.size; i++)
                   g_ref.data[i] = fasta[gnr].ref.data[i];

               new_ref = va_allocVecU8(g_ref.size + g_ins_vec.size);
               for (i = 0; i < new_ref.size; i++)
                   new_ref.data[i] = 6;

               for (i = 0; i < g_ins_vec.size; i++)
               {
                   VA_ASSERT((g_ins_vec.data[i] <= new_ref.size) &&
                             (g_ins_vec.data[i] > 0));
                   new_ref.data[g_ins_vec.data[i] - 1] = 4;
               }

               j = 0;
               for (i = 0; i < new_ref.size; i++)
               {
                   if (new_ref.data[i] != 4)
                       new_ref.data[i] = g_ref.data[j++];
                   VA_ASSERT(new_ref.data[i] != 6);
               }

               free(g_ref.data);
               g_ref = new_ref;
            }            

            for (i = 0; i < g_ins_vec.size; i++)
                g_ins_vec.data[i] -= i + 1;

            /* align reads */
            g_reads = align_reads(g_ref, g_reads, g_ins_vec);

            for (i = 0; i < g_ins_vec.size; i++)
                g_ins_vec.data[i] += i + 1;

            /* prepare the insertion vector correctly for the reference */
            if (g_ins_vec.size > 0)
            {
                VA_VecI32 g_ins_vec2 = va_allocVecI32(g_ins_vec.size);

                for (j = 0; j < g_ins_vec2.size; j++)
                    g_ins_vec2.data[j] = g_ins_vec2.data[j] + j + 1;

                ins_vec_add = va_unique(g_ins_vec2);
                free(g_ins_vec2.data);
            }
            else
            {
                memset(&ins_vec_add, 0, sizeof(ins_vec_add));
            }

            {
                ReadsNode_t * rd = (ReadsNode_t *)g_reads.data;
                for (j = 0; j < g_reads.size; j++)
                    g_readstarts.data[j] = rd[j].chrstart;
            }

            /* create and populate the cdata structure */
            cdata = real_cdata_prep(FALSE, TRUE);
            cdata.ref = g_ref;
            cdata.reads = g_reads;
            cdata.readstarts = g_readstarts;
            cdata.read_size = 35;
            cdata.useqvals = FALSE;

            /* run the VARiD Algorithm and get initial predictions */

            /* varid algorithm */
            varid(cdata, 0, &maps, &snpgroup.data[gnr].covmap, &p,useSam);

            predict(p, maps, &snpgroup.data[gnr].maxvals_2, 
                    &snpgroup.data[gnr].lmat, 
                    &snpgroup.data[gnr].cmat);

            if (p.data != NULL)
                free(p.data);
 
            /* Fix SNPs TODO: use sparse?
             * destroy_state doesn'yt use thresholds anymore...
             */
            destroy_state(cdata, snpgroup.data[gnr].cmat, g_reads, cdata.read_size, 
                          g_readstarts, snpgroup.data[gnr].covmap, thr_vec_snp, 
                          snpgroup.data[gnr].lmat, &snpgroup.data[gnr].ch_list, 
                          &tot, &snpgroup.data[gnr].top, useSam);

            snpgroup.data[gnr].ref = g_ref;
            snpgroup.data[gnr].ins_vec_add = ins_vec_add;
            snpgroup.data[gnr].readstarts = g_readstarts;

            if (maps.full_letter.data != NULL)
            {
                free(maps.full_letter.data);
                maps.full_letter.data = NULL;
            }

            if (maps.call_letter.data != NULL)
            {
                free(maps.call_letter.data);
                maps.call_letter.data = NULL;
            }

            if (maps.call_color.data != NULL)
            {
                free(maps.call_color.data);
                maps.call_color.data = NULL;
            }

            rd  = (ReadsNode_t *)cdata.reads.data;
            for (j = 0; j < cdata.reads.size; j++)
            {
                if (rd[j].name != NULL)
                    free(rd[j].name);
#if 0
                if (rd[j].letseq.data != NULL)
                    free(rd[j].letseq.data);
#endif
                if (rd[j].colseq.data != NULL)
                    free(rd[j].colseq.data);
                if (rd[j].ins.data != NULL)
                    free(rd[j].ins.data);
            }
            free(rd);

            if (cdata.readstarts.data != NULL)
                free(cdata.readstarts.data);

            {
                VA_VecU8 hap;
                VA_VecU8 hap2;
                VA_VecU8 pred1_whole;
                VA_VecU8 pred2_whole;

                memset(&hap, 0, sizeof(hap));

                get_pred(snpgroup.data[gnr].covmap, maps,
                         snpgroup.data[gnr].ch_list, thr_vec_p,
                         snpgroup.data[gnr].maxvals_2, snpgroup.data[gnr].lmat,
                         hap, hap, &pred1_whole, &pred2_whole);

                writePrettybase(snpgroup.data[gnr].range.data[0],
                            fasta[gnr].hdr, don_name, pred1_whole, pred2_whole,
                            snpgroup.data[gnr].ref,
                            wrPrettyFile);
            }
        }
    }

    rc = clock_gettime(CLOCK_REALTIME, &te);
    VA_ASSERT(rc == 0);

    tdiff = va_TimeSub(&te, &mastertic);
    fprintf(stderr, "Time for conputation: %ld.%03ld secs\n",
            tdiff.tv_sec, tdiff.tv_nsec/NS_IN_MS);
}

static void initThresholdVec(
    int            doEval,
    VA_VecDouble * thr_snp,
    VA_VecDouble * thr_p)
{
    VA_ASSERT(thr_snp != NULL);
    VA_ASSERT(thr_p != NULL);

    if (doEval)
    {
        *thr_snp = va_allocVecDouble(1);
        *thr_p = va_allocVecDouble(1);

        thr_snp->data[0] = 0.4;
        thr_p->data[0] = 0.98;
    }
    else
    {
        *thr_snp = va_allocVecDouble(1);
        *thr_p = va_allocVecDouble(1);

        thr_snp->data[0] = 0.1;
        thr_p->data[0] = 0.999;
    }
}

static char ** initGroupFiles(
    const char * readsfile,
    const char * path,
    int          nfiles)
{
    int i;
    char ** filenames;
    char buf[BUFSIZ];

    VA_ASSERT(readsfile != NULL);
    VA_ASSERT(path != NULL);

    VA_ASSERT(nfiles > 0);

    filenames = (char **)malloc(sizeof(char *) * nfiles);
    VA_ASSERT(filenames != NULL);


    for (i = 0; i < nfiles; i++)
    {
        char * p = va_basename(readsfile);

        buf[0] = 0;
        sprintf(buf, "%s/%s.group%d", path, p, i + 1);
        free(p);
        filenames[i] = (char *)malloc(strlen(buf) + 1);
        VA_ASSERT(filenames[i] != NULL);
    
        strcpy(filenames[i], buf);
    }

    return (filenames);
}

static VA_Vector getFastaReference(
    const char * reffile) 
{
    VA_Vector fasta;
    int i;
    int j;
    Fasta_t * p;
    int ret;

    VA_ASSERT(reffile != NULL);

    ret = va_readfasta(reffile, &fasta);
    VA_ASSERT(ret == 0);

    p = (Fasta_t *)fasta.data;
    for (i = 0; i < fasta.size; i++)
    {
        VA_VecU8 v;

        p[i].ref = va_allocVecU8(p[i].seq.size - 1);

        for (j = 0; j < p[i].ref.size; j++)
            p[i].ref.data[j] = tolower(p[i].seq.data[j]);
 
        v = nuc2nuc(p[i].ref);
        free(p[i].ref.data);
        memcpy(&p[i].ref, &v, sizeof(v));
        fprintf(stderr, "\t %20s:%d nt \n", p[i].hdr, p[i].ref.size);
    }

    return (fasta);
}

static int cmpSortI32(
   const void * rec1,
   const void * rec2)
{
   VA_INT32 p1;
   VA_INT32 p2;

   VA_ASSERT((rec1 != NULL) && (rec2 != NULL));

   p1 = *(VA_INT32 *)rec1;
   p2 = *(VA_INT32 *)rec2;

   return (p1 - p2);
}


