/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include "va_assert.h"
#include "va_log.h"
#include "va_time.h"
#include "varid.h"

/******************************************************************************
* aligns the mapped reads and the reference
*   [new_ref, reads] = ALIGN_READS(ref, reads, ins_vec) aligns the mapped
*   reads and the reference in the very rare case of a gap present in the
*   mapped read and not present in the reference or another read at that
*   position.
*
*   Parameters:
*   ref - the reference in ns letterspace
*   reads - the read vector struct
*   ins_vector - a vector indicating places in the reference where an
*       insertion occurs due to a mapping
*
*   Returns:
*   new_ref - the aligned reference
*   reads - the aligned reads
*
*   Author: Adrian V Dalca
******************************************************************************/
VA_Vector align_reads(
    VA_VecU8  ref,
    VA_Vector reads,
    VA_VecI32 ins_vec)
{    
    int i;
    int rc;
    ReadsNode_t * rd = (ReadsNode_t *)reads.data;
    struct timespec areadtic;
    struct timespec te;
    struct timespec tdiff;
    
    /******************
     * New method
     ******************/

    fprintf(stderr, "Aligning reads ...\n");

    /* just return if there is no work to be done */
    if (ins_vec.size > 0)
    {
        rc = clock_gettime(CLOCK_REALTIME, &areadtic);
        VA_ASSERT(rc == 0);

        for (i = 0; i < reads.size; i++)
        { 
            VA_VecI32 f;
            int j;
            VA_INT32 delay;
            VA_VecU8 seq;
            VA_VecU8 seq_wo_ins;
            VA_VecU8 new_seq;

            /* decide what type of support we are dealing with: 
             * letter of colors 
             */
            int isLetterspace = (rd[i].seqtype == LETTERSPACE_READ) ? 
                                TRUE : FALSE;

            /* get any insertions that are between [s, fi)
             * these bounds are chosen specifically so because we care about
             * an insertion @ s (since it would be put after s) but not about
             * one @ fi (since it would be put AFTER fi, and hence not in our
             * care
             */
            VA_INT32 s = rd[i].chrstart;
            VA_INT32 fi = rd[i].chrend;
            VA_INT32 fi_wo_ins = fi - rd[i].ins.size;
            VA_VecI32 ins_local;
            VA_VecI32 ins_local2;
            VA_VecI32 readins;
            int snp_incr = 0;

            int isForward = (rd[i].strand == '+') ? TRUE : FALSE;

            if (isLetterspace)
                seq = rd[i].letseq;
            else
                seq = rd[i].colseq;

            memset(&f, 0, sizeof(f));
            memset(&new_seq, 0, sizeof(new_seq));
            memset(&seq_wo_ins, 0, sizeof(seq_wo_ins));
            memset(&ins_local2, 0, sizeof(ins_local2));
            memset(&ins_local, 0, sizeof(ins_local));

            if (ins_vec.size > 0)
            {
                f = va_allocVecI32(ins_vec.size);
                f.size = 0;
                
                for (j = 0; j < ins_vec.size; j++)
                {
                    if ((ins_vec.data[j] >= s) && (ins_vec.data[j] < fi_wo_ins))
                  
                        f.data[f.size++] = j + 1;
                }
                VA_ASSERT(f.size <= ins_vec.size);
            }

            if (f.size > 0)
            {
                VA_VecI32 readins_del;
                int k;

                ins_local = va_allocVecI32(f.size);
                for (j = 0; j < ins_local.size; j++)
                {
                    VA_ASSERT((f.data[j] > 0) && (f.data[j] <= ins_vec.size));
                    ins_local.data[j] = ins_vec.data[f.data[j] - 1] - rd[i].chrstart + 1;
                    if (isForward && !isLetterspace)
                        ins_local.data[j]--;
                }
 
                memset(&readins_del, 0, sizeof(readins_del));
                memset(&readins, 0, sizeof(readins));

                if (rd[i].ins.size > 0)
                {
                    readins = va_allocVecI32(rd[i].ins.size);
                    readins_del = va_allocVecI32(rd[i].ins.size);

                    memcpy(readins.data, rd[i].ins.data, 
                        rd[i].ins.size * sizeof(VA_INT32));
                    for (j = 0; j < rd[i].ins.size; j++)
                        readins_del.data[j] = rd[i].ins.data[j] + j + 1;
                }                    

                if (readins.size > 0)
                {
                    int sum = 0;
                    
                    for (j = 0; j < readins.size; j++)
                        sum += (readins.data[j] > ins_vec.data[ins_vec.size - 1] - 
                               rd[i].chrstart + 1) ? 1 : 0;
                    if (sum > 0)
                    {
                        size_t oldsz = ins_local.size;
                        k = 0;
                        ins_local = va_reallocVecI32(ins_local, ins_local.size + sum);
                        for (j = 0; j < readins.size; j++)
                        {
                            VA_ASSERT(ins_vec.size > 0);
                            if (readins.data[j] > ins_vec.data[ins_vec.size - 1] - 
                                                  rd[i].chrstart + 1)
                            {
                                ins_local.data[oldsz + k++] = rd[i].ins.data[j];
                                VA_ASSERT(oldsz + k <= ins_local.size);
                            }
                        }
                    }    
                }

                VA_ASSERT(seq.size - readins_del.size > 0);
                seq_wo_ins = va_allocVecU8(seq.size - readins_del.size);
                k = 0;
                for (j = 0; j < seq.size; j++)
                {
                    int jj = 0;
                    int toskip = FALSE;

                    for (jj = 0; jj < readins_del.size; jj++)
                    {
                        if (j + 1 == readins_del.data[jj])
                        {
                            toskip = TRUE;
                            break;
                        }
                    }
                        
                    if (toskip == FALSE)
                        seq_wo_ins.data[k++] = seq.data[j];
                }
                VA_ASSERT(k == seq_wo_ins.size);

                new_seq = va_allocVecU8(seq.size - readins.size + ins_local.size);
                for (j = 0; j < new_seq.size; j++)
                    new_seq.data[j] = 6;

                ins_local2 = va_allocVecI32(ins_local.size);
                for (j = 0; j < ins_local2.size; j++)
                {
                    ins_local2.data[j] = ins_local.data[j] + j + 1;
                    VA_ASSERT((ins_local2.data[j] > 0) && 
                              (ins_local2.data[j] <= new_seq.size));
                    new_seq.data[ins_local2.data[j] - 1] = 4;
                }
                   
                k = 0;
                for (j = 0; j < new_seq.size; j++)
                {
                    if (new_seq.data[j] == 6)
                        new_seq.data[j] = seq_wo_ins.data[k++];
                }

                if (readins.size > 0)
                {
                    VA_VecI32 o;
                    VA_VecI32 u;
                    VA_VecI32 orig_ins;

                    orig_ins = va_allocVecI32(readins.size);
                    o = va_allocVecI32(orig_ins.size);

                    memcpy(orig_ins.data, readins.data, 
                        sizeof(VA_INT32) * orig_ins.size);
                  
                    memcpy(o.data, orig_ins.data, 
                        sizeof(VA_INT32) * o.size);
                    u = va_unique(orig_ins);
                    for (j = 0; j < u.size; j++)
                    {
                        VA_VecI32 osf = va_allocVecI32(ins_local.size);
                        VA_VecI32 osl = va_allocVecI32(o.size);

                        osf.size = 0;
                        osl.size = 0;
                        for (k = 0; k < ins_local.size; k++)
                        {
                            if (u.data[j] == ins_local.data[k])
                                osf.data[osf.size++] = k + 1;
                        }

                        for (k = 0; k < o.size; k++)
                        {
                            if (u.data[j] == o.data[k])
                                osl.data[osl.size++] = k + 1;
                        }

                        VA_ASSERT(osl.size <= o.size);
                        VA_ASSERT(osf.size >= osl.size);

                        for (k = 0; k < osl.size; k++)
                        {
                            VA_ASSERT((osl.data[k] > 0) && 
                                      (osl.data[k] - 1 <= orig_ins.size));
                            orig_ins.data[osl.data[k] - 1] += osf.data[k];
                        }

                        free(osf.data);
                        free(osl.data);
                    }

                    for (k = 0; k < orig_ins.size; k++)
                    {
                        VA_ASSERT(new_seq.data[orig_ins.data[k] - 1] == 4);
                        new_seq.data[orig_ins.data[k] - 1] =
                            seq.data[readins_del.data[k] - 1];
                    }

                    if (u.data != NULL)
                        free(u.data);
                    if (o.data != NULL)
                        free(o.data);
                    if (orig_ins.data != NULL)
                        free(orig_ins.data);
                }

                if (isLetterspace)
                {
                    if (rd[i].letseq.data != NULL)
                        free(rd[i].letseq.data); 
                    rd[i].letseq = new_seq;
                }
                else
                {
                    if (rd[i].colseq.data != NULL)
                        free(rd[i].colseq.data); 
                    rd[i].colseq = new_seq;
                }

                snp_incr = f.size - readins.size;
            }

            /* update the starting position */
            delay = 0;
            for (j = 0; j < ins_vec.size; j++)
                if (ins_vec.data[j] < rd[i].chrstart)
                    delay++;

            rd[i].chrstart += delay;
            rd[i].chrend += delay + snp_incr;

            if (ins_local.data != NULL)
                free(ins_local.data);
            if (ins_local2.data != NULL)
                free(ins_local2.data);
            if (seq_wo_ins.data != NULL)
                free(seq_wo_ins.data);
            if (f.data != NULL)
                free(f.data);
        }

        rc = clock_gettime(CLOCK_REALTIME, &te);
        VA_ASSERT(rc == 0);

        tdiff = va_TimeSub(&te, &areadtic);

        /* done read prep */
        fprintf(stderr, "\t Done in: %ld.%03ld secs\n",
            tdiff.tv_sec, tdiff.tv_nsec/NS_IN_MS);
    }

    return (reads);
}


