/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "va_assert.h"
#include "va_log.h"
#include "varid.h"

static int compareReads(
    const void * rec1,
    const void * rec2);

static int compareInsvec(
    const void * rec1,
    const void * rec2);

static VA_Vector readsDup(
    VA_Vector  reads,
    VA_UINT32  n);

static VA_UINT32 realloc_reads(
    VA_UINT32      readnr,
    VA_UINT32      maxreads,
    ReadsNode_t *  in_reads,
    ReadsNode_t ** out_reads,
    VA_INT32    *  in_readstarts,
    VA_INT32    ** out_readstarts);

static VA_VecI32 concatenate_ins(
    VA_VecI32 ins,
    VA_VecI32 toappend);

static VA_INT32 * readstarts = NULL;

/******************************************************************************
*   extract the reads and populates the necessary struct vector
*   [reads, insvec] = GET_READS(mapfile) will read the mapping in file
*       mapfile and populate the struct vector reads, as well as the
*       insertion vector insvec.
*
*   [reads, insvec] = GET_READS(mapfile, nr_reads, nr_lines) will proceed
*   similarly, however by supplying the number of reads and the number of
*   lines we can achieve better storage and get a waitbar.
*
*   Parameters:
*   mapfile - path to the mapped reads file (probcalc format)
*
*   Optional:
*   nr_reads - if given, will be able to estimate struct sizes correctly.
*   nr_lines - if given, can run waitbar
*
*   Returns:
*   reads  - a (nr_reads) vector of structs with the fields indicated in
*       varid_const.m
*
*   Author: Adrian V. Dalca
*******************************************************************************/
void get_reads(
    char       * mapfile, 
    VA_UINT32    nr_reads, 
    VA_UINT32    nr_lines,
    VA_Vector  * outreads,
    VA_VecI32  * outinsvec)
{
	//mapfile = "test.sam.300";
    FILE * fid;
    ReadsNode_t * reads;
    char line[BUFSIZ];
    VA_INT32 readnr = 0;
    int i;
    CurrReads_t cur_read;
    cur_read.Qline[0] = 0;
    VA_VecI32 insvec;
    VA_UINT32 maxreads = (nr_reads > 0) ? nr_reads : 10000; 

    VA_ASSERT(outreads != NULL);
    VA_ASSERT(outinsvec != NULL);

    memset(outreads, 0, sizeof(outreads));
    memset(outinsvec, 0, sizeof(outinsvec));

    reads = (ReadsNode_t *)malloc(sizeof(ReadsNode_t) * maxreads);
    VA_ASSERT(reads != NULL);
    memset(reads, 0, sizeof(ReadsNode_t) * maxreads);

    readstarts = (VA_INT32 *)malloc(sizeof(VA_INT32) * maxreads);
    VA_ASSERT(reads != NULL);
    memset(readstarts, 0, sizeof(VA_INT32) * maxreads);
    memset(&insvec, 0, sizeof(insvec));

    /* record all readstarts. Also acts as a flag for reads that are skipped */
    
    /* Start doing the work */
    fid = fopen(mapfile, "r");
    VA_ASSERT(fid != NULL);
 
    readnr = 0;
    i = 0;
    while (1)
    {
        if ((nr_lines > 0) && (i > nr_lines))
            break;

        /* get the line */
        if (fgets(line, BUFSIZ, fid) == NULL)
        {
            if ((nr_lines >= 0) && (i == nr_lines))
                break;

            /*LOG_ERROR3("Unexpected EOF; %s: current/expected lines = %d/%d",
                mapfile, i, nr_lines);
            exit(EXIT_FAILURE);*/
        }
        else
        {
            switch (line[0])
            {
            case '>':
                {
                    char * p = line;

                    /* process previous read (if this is not the first line/read */
                    if (readnr > 0)
                    {
                        VA_UINT32 pskipread = 0;

                        if (readnr > maxreads)
                        {
                            maxreads = realloc_reads(readnr, maxreads, reads, &reads, 
                                           readstarts, &readstarts);
                        }
                  
                        process_read(&cur_read, &reads[readnr - 1], &readstarts[readnr - 1],
                            &pskipread, &insvec);

                        /* copy the new reads */
                        if (pskipread != 0)
                            readnr -= 1;
                    }

                    VA_ASSERT(strlen(line) < sizeof(line));
                    readnr++;
                    strcpy(cur_read.sline, line);
                    cur_read.Qline[0] = 0;
                }   
                break;
            case 'G':
                strcpy(cur_read.Gline, line);
                break;
            case 'T':
                strcpy(cur_read.Tline, line);
                break;
            case 'R':
                strcpy(cur_read.Rline, line);
                break;
            case 'Q':
                strcpy(cur_read.Qline, line);
                break;
            default:
                break;
            }
        }
        i++;
    }

    if (readnr > 0)
    {
        VA_UINT32 pskipread;

        if (readnr > maxreads)
        {
           maxreads = realloc_reads(readnr, maxreads, reads, &reads,
                          readstarts, &readstarts);
        }

        process_read(&cur_read, &reads[readnr - 1], &readstarts[readnr - 1],
            &pskipread, &insvec);

        if (pskipread != 0)
            readnr -= 1;
    }

    if (readnr > 0)
    {
        /* realloc with correct readnr elements */
        VA_Vector in_reads;
        VA_Vector out_reads;
        in_reads.data = reads;
        in_reads.size = readnr;

        out_reads = readsDup(in_reads, readnr);
        free(in_reads.data);
        reads = out_reads.data;
    }

    /* extract the good reads */
    /* only take out the good reads (reads within the constraints */
    if (readnr > 0)
    {
        int good_nr_reads = 0;
        VA_INT32 * good_reads_idx = NULL;
        ReadsNode_t * new_reads = NULL;
        
        i = 0;
        while (i < nr_reads)
        {
           if (readstarts[i] > 0)
           {
              if (good_nr_reads < i)
              {
                  memcpy(&reads[good_nr_reads], &reads[i], sizeof(ReadsNode_t));
                  readstarts[good_nr_reads] = readstarts[i];
              }
              good_nr_reads++;
           }
           i++;
        }

        VA_ASSERT(good_nr_reads > 0);
        good_reads_idx = (VA_INT32 *)malloc(sizeof(VA_INT32) * good_nr_reads);
        VA_ASSERT(good_reads_idx != NULL);

        for (i = 0; i < good_nr_reads; i++)
            good_reads_idx[i] = i;

        /* sort the reads */
        qsort(good_reads_idx, good_nr_reads, sizeof(VA_INT32), compareReads); 
        
        new_reads = (ReadsNode_t *)malloc(sizeof(ReadsNode_t) * good_nr_reads);
        VA_ASSERT(new_reads != NULL);
        for (i = 0; i < good_nr_reads; i++)
            memcpy(&new_reads[i], &reads[good_reads_idx[i]], sizeof(ReadsNode_t));

        if (insvec.size > 0)
        {
            qsort(insvec.data, insvec.size, sizeof(VA_INT32), compareInsvec); 
            outinsvec->data = insvec.data;
            outinsvec->size = insvec.size;

        }
        outreads->data = new_reads;
        outreads->size = good_nr_reads;
        free(reads);
    }
    else
    {
        free(reads);
        free(readstarts);
    }

    fclose(fid);
}

VA_Vector readsDup(
    VA_Vector  reads,
    VA_UINT32  n)
{
    ReadsNode_t * p = (ReadsNode_t *)reads.data;
    ReadsNode_t * q;
    VA_Vector ret;
    int i;
  
    VA_ASSERT((reads.size > 0) && (reads.data != NULL));
    VA_ASSERT(n >= reads.size);

    q = (ReadsNode_t *)malloc(sizeof(ReadsNode_t) * n);
    VA_ASSERT(q != NULL);

    for (i = 0; i < reads.size; i++)
    {
        q[i].name = p[i].name;
        if (p[i].name != NULL)
            p[i].name = NULL;

        q[i].strand = p[i].strand;
        q[i].chrstart = p[i].chrstart;
        q[i].chrend = p[i].chrend;
        q[i].extras.score = p[i].extras.score;
        q[i].extras.nodds = p[i].extras.nodds;
        q[i].extras.pgenome = p[i].extras.pgenome;
        q[i].extras.pchance = p[i].extras.pchance;
        q[i].extras.haplome = p[i].extras.haplome;
        q[i].seqtype = p[i].seqtype;

        q[i].letseq.size = p[i].letseq.size;
        q[i].letseq.data = p[i].letseq.data;
        if (p[i].letseq.data != NULL)
            p[i].letseq.data = NULL;
        q[i].colseq.size = p[i].colseq.size;
        q[i].colseq.data = p[i].colseq.data;
        if (p[i].colseq.data != NULL)
            p[i].colseq.data = NULL;
        q[i].qvals.size = p[i].qvals.size;
        q[i].qvals.data = p[i].qvals.data;
        if (p[i].qvals.data != NULL)
            p[i].qvals.data = NULL;

        q[i].ins.size = p[i].ins.size;
        q[i].ins.data = p[i].ins.data;
        if (p[i].ins.data != NULL)
            p[i].ins.data = NULL;
    }

    ret.size = n;
    ret.data = q;

    return (ret);
}

static int compareReads(
    const void * rec1,
    const void * rec2)
{
    VA_INT32 p1;
    VA_INT32 p2;

    VA_ASSERT((rec1 != NULL) && (rec2 != NULL));

    p1 = *(VA_INT32 *)rec1;
    p2 = *(VA_INT32 *)rec2;

    return (readstarts[p1] - readstarts[p2]);
}

static int compareInsvec(
    const void * rec1,
    const void * rec2)
{
    VA_INT32 p1;
    VA_INT32 p2;

    VA_ASSERT((rec1 != NULL) && (rec2 != NULL));

    p1 = *(VA_INT32 *)rec1;
    p2 = *(VA_INT32 *)rec2;

    return (p1 - p2);
}
        
static VA_UINT32 realloc_reads(
    VA_UINT32      readnr,
    VA_UINT32      maxreads,
    ReadsNode_t *  in_reads,
    ReadsNode_t ** out_reads, 
    VA_INT32    *  in_readstarts,
    VA_INT32    ** out_readstarts)
{
    VA_Vector vin_reads;
    VA_Vector vout_reads;
    VA_INT32 * tmp_readstarts;

    vin_reads.data = in_reads;
    vin_reads.size = maxreads;

    /* realloc reads */
    maxreads *= 2;
    vout_reads = readsDup(vin_reads, maxreads);
    free(vin_reads.data);
    *out_reads = vout_reads.data;

    /* realloc readstarts */
    tmp_readstarts = (VA_INT32 *)malloc(sizeof(VA_INT32) * maxreads);
    VA_ASSERT(tmp_readstarts != NULL);
    memcpy(tmp_readstarts, readstarts, sizeof(VA_INT32) * readnr);
    free(in_readstarts);
    *out_readstarts = tmp_readstarts;
    return (maxreads);
}

static VA_VecI32 concatenate_ins(
    VA_VecI32 ins,
    VA_VecI32 toappend)
{
    VA_VecI32 retins;
  
    retins.size = ins.size;
    retins.data = ins.data;

    /* concatenate insvec */
    if (toappend.size > 0)
    {
        int i;
        int j;
        int n = ins.size;
        int flag = FALSE;
        VA_VecI32 pins = va_allocVecI32(ins.size + toappend.size);
        if (n > 0)
        {
            VA_ASSERT(n <= pins.size);
            memcpy(pins.data, ins.data, n * sizeof(VA_INT32));
            free(ins.data);
        }

        for (i = 0; i < toappend.size; i++)
        {
            flag = FALSE;
            for (j = 0; j < ins.size; j++)
            {
                if (toappend.data[i] == pins.data[j])
                {
                    flag = TRUE;
                    break; 
                }
            }
            if (flag == FALSE)
                pins.data[n++] = toappend.data[i];
        }
                                
        if (pins.size != n)
            pins = va_reallocVecI32(pins, n);

        free(toappend.data);
        memcpy(&retins, &pins, sizeof(pins));
    }

    return (retins);
}

