/*
 * VARiD: A Variation Detection Framework for Colorspace and Letterspace platforms
 * Contact: <varid at cs dot toronto dot edu>
 * Copyright 2010 the authors (see README).
 *
 * This file is part of VARiD.
 *
 * VARiD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * VARiD is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with VARiD.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include "va_assert.h"
#include "va_log.h"
#include "varid.h"
#include <errno.h>

void uppercase(char* s) {
	while(*s!='\0') {
		*s=toupper(*s);
		s++;
	}
}


void cs_insert_gaps(char* cs,char* lsa, char* temp,int start, int end, int readlength) {
	int temp_index=0;
	int cs_index=0;
	int lsa_index=0;
	VA_ASSERT(cs[0] == 'T' || cs[0] == 'A' || cs[0] == 'C' || cs[0] == 'G');
	temp[temp_index++]=cs[cs_index++];
	//append if need be
	int prefix;
	for (prefix=start; prefix>1; prefix--) {
		temp[temp_index++]=cs[cs_index++];
		VA_ASSERT(lsa[lsa_index++]=='-');
	} 
	//get the middle
	int middle=end-start+1;
	while ( middle>0 ) {
		if (lsa[lsa_index]=='-') {
			temp[temp_index++]=lsa[lsa_index++];
		} else {
			temp[temp_index++]=cs[cs_index++];
			lsa_index++;
			middle--;
		}
	}
	//get the suffix
	int suffix=readlength-end; 	
	while ( suffix > 0 ) {
		temp[temp_index++]=cs[cs_index++];
		suffix--;
	}
	temp[temp_index]='\0';
	strcpy(cs,temp);
}

void reverse(char* s, char* t) {
	int l=strlen(s);
	int i;
	for (i=0; i<l; i++) {
		switch (s[i]) {
			case 'A':
				t[l-i-1]='T';
				break;
			case 'a':
				t[l-i-1]='t';
				break;
			case 'C':
				t[l-i-1]='G';
				break;
			case 'c':
				t[l-i-1]='g';
				break;
			case 'G':
				t[l-i-1]='C';
				break;
			case 'g':
				t[l-i-1]='c';
				break;
			case 'T':
				t[l-i-1]='A';
				break;
			case 't':
				t[l-i-1]='a';
				break;
			case 'N':
				t[l-i-1]='N';
				break;
			case 'n':
				t[l-i-1]='n';
				break;
			case '-':
				t[l-i-1]='-';
				break;
			default:
				fprintf(stderr,"There has been a error in getting reverse complement of %s\n",s);
				exit(1);
		}
	}
	t[l]='\0';
	strcpy(s,t);
}


void split_reads(
		const char  * readsfile,
		char       ** filenames,
		SnpGroups_t * snpgroup,
		VA_UINT8      method,
		VA_INT32   ** fid_lines,
		VA_INT32   ** fid_reads,
		int			  useSam,
		VA_Vector     fastaVec)
{
	FILE ** fids = NULL;
	FILE * fid = NULL;
	int stderr_grp = snpgroup->size;
	int write_grp;
	int i;
	int j;
	char line[BUFSIZ];
	//char * scan_middle = "%s %s %c %d %d %d %d %d %d %s %lf %lf %lf";
	char * scan_middle = "%s\t%s\t%c\t%d\t%d\t%d\t%d\t%d\t%d\t%s\t%lf\t%lf\t%lf";
	VA_INT32 * f_lines = NULL;
	VA_INT32 * f_reads = NULL;
	int totreads = 0;
	int usedreads = 0;
	char qname[BUFSIZ], rname[BUFSIZ], cigar[BUFSIZ], mrnm[BUFSIZ], seq[BUFSIZ], qual[BUFSIZ];
	char tempreadstring[BUFSIZ],readstring[BUFSIZ], genomestring[BUFSIZ];
	int flag, mapq, mpos, isize, usequal;
	int pos;
	VA_INT32 leftmost;
	VA_INT32 rightmost;
	VA_INT32 readstart;
	VA_INT32 readend;
	double normodds;
	double pgenome;
	double pchance;
	VA_INT32 readlength;
	VA_INT32 score;
	char editstring[BUFSIZ];
	char readname[BUFSIZ];
	char contigname[BUFSIZ];
	char strand='*';
	char color_space_read[BUFSIZ];
	int color_space=FALSE;
	char * p = line;
	char * chr;
	int chr2;
	int refstart = 0;

	VA_ASSERT(fid_lines != NULL);
	VA_ASSERT(fid_reads != NULL);

	fids = (FILE **)malloc(sizeof(FILE *) * (snpgroup->size + 1));
	VA_ASSERT(fids != NULL);

	/* PRECONDITION: snp groups don't overlap in ranges */

	fid = fopen(readsfile, "r");
	VA_ASSERT(fid != NULL);

	/* open read files */
	f_lines = (VA_INT32 *)malloc(sizeof(VA_INT32) * (snpgroup->size + 1));
	VA_ASSERT(f_lines != NULL);
	memset(f_lines, 0, (snpgroup->size + 1) * sizeof(VA_INT32));

	f_reads = (VA_INT32 *)malloc(sizeof(VA_INT32) * (snpgroup->size + 1));
	VA_ASSERT(f_reads != NULL);
	memset(f_reads, 0, (snpgroup->size + 1) * sizeof(VA_INT32));

	for (i = 0; i < snpgroup->size; i++)
	{
		if (filenames[i] != NULL)
		{
			fids[i] = fopen(filenames[i], "w");
			if (fids[i]==NULL) {
				perror("FATAL IO Error");
				exit(1);
			}
			VA_ASSERT(fids[i] != NULL);
		}
		else
		{
			fids[i] = NULL;
		}
	}

	fids[stderr_grp] = stderr;
	VA_ASSERT(fids[stderr_grp] != NULL);

	/* the writing index. */
	write_grp = stderr_grp;

	totreads = 0;
	usedreads = 0;

	int clipped_or_skipped=0;

	/* go through each line and re-write it in the right file */
	/* get the line (with newline char) */
	while (fgets(line, BUFSIZ, fid) != NULL)
	{
		/* if this is a new line */
		if(useSam == 1){
			if (line[0] == '@'){
				continue;
			}
			totreads++;
			color_space=FALSE;
			
			char* save_pointer;
			char* str=line;
			char* tok;
			int parsed_tokens=0;
			char* cs;
			//char *strtok_r(char *str, const char *delim, char **saveptr);
			while ( (tok=strtok_r(str,"\t",&save_pointer))!=NULL) {
			//int res= sscanf(end_read_name+1,"%d\t%s\t%d\t%d\t%s\t%s\t%d\t%d\t%s\t%s",
			//		&flag, rname, &pos, &mapq,
			//		cigar, mrnm, &mpos, &isize, seq, qual);
				switch (parsed_tokens) {
					case 0: strcpy(qname,tok); break;
					case 1:	flag=atoi(tok); break;
					case 2:	strcpy(rname,tok); break;
					case 3: pos=atoi(tok); break;
					case 4: mapq=atoi(tok); break;
					case 5: strcpy(cigar,tok); break;
					case 6:	strcpy(mrnm,tok); break;
					case 7:	mpos=atoi(tok); break;
					case 8:	isize=atoi(tok); break;
					case 9: strcpy(seq,tok); break;
					case 10: strcpy(qual,tok); break;
					default:
						cs=strstr(tok,"CS:Z:");
						if (cs!=NULL) {
							color_space=TRUE;
							cs+=5;
							sscanf(cs,"%s",color_space_read);
						}
						break;
				} 
				parsed_tokens++;
				str=NULL;
			}
			if (parsed_tokens<10) {
				fprintf(stderr,"Skipping read! failed parse\n");
				continue;
			}
			if (pos<0) {
				fprintf(stderr,"Skipping read! invalid position\n");
				continue;
			}
			if (*qual == '*'){
				usequal = 0;
			}

			
			//fprintf(stderr,"Refrence <%s>\n",rname);
			strncpy(contigname, rname, BUFSIZ);
			//fprintf(stdout,"contig name '%s'\n",contigname);
			leftmost = pos;
			readlength = strlen(seq);

			//calculate rightmost and build read
			// parse cigar string
			char *c = cigar;
			char *cread = readstring;
			char *cgenome = genomestring;
			char *curread = seq;
			char *curgenome;

			if (flag & 16) {
				strand='-';
			} else {
				strand='+';
			}
 

			Fasta_t *refs = (Fasta_t *)fastaVec.data;
			int s = fastaVec.size;

			//find the right refrence sequence
			int i;
			for(i = 0; i < s;i++){
				if (strcmp(refs[i].hdr,rname) == 0){
					break;
				}
			}
			if (i == s){
				//fprintf(stderr,"Refrence <%s> not found\n",rname);
				//fprintf(stderr,"line:%s",line);
				//exit(1);
				continue;
			}

			curgenome = refs[i].seq.data + leftmost - 1; //need 0 basd
			//fprintf(stdout,"matched contig '%s'\n",refs[i].hdr);
			readstart=1;
			readend=readlength;
			int inserts=0;
			int deletions=0;
			int count = 0;
			int skip=0;
			int start = TRUE;
			while (*c != '\0'){
				if (*c <= '9' && *c >= '0'){
					count = count*10 + *c - '0';
				} else if (*c == 'S' || *c == 'H') {
					memset(cread,'-',count);
					cread+=count;
					if (*c == 'S') {
						curread+=count;
					} else {
						readlength+=count;
					}
					if(start){
						if (strand=='+') {
							readstart=count+1;
							if (*c =='H') {
								readend+=count;
							}
						} else {
							if (*c == 'S') {
								readend-=count;
							}
						}
						if (count>(leftmost-1)) {
							memset(cgenome,'-',count-leftmost+1);
							cgenome+=count-leftmost+1;
							count=leftmost-1;	
						}
						
						strncpy(cgenome,curgenome-count,count);
						cgenome+=count;
						start = FALSE;
					} else {
						strncpy(cgenome,curgenome,count);
						int x;
						for (x=0; (cgenome[x]!='\0' && x<count); x++);
						if ((count-x)>0) {
							cgenome+=x;
							memset(cgenome,'-',count-x);
							cgenome+=count-x;
						} else {
							cgenome+=count;
						}
						if (strand=='+') {
							if (*c == 'S') {
								readend-=count;
							}
						} else {
							readstart=count+1;
							if (*c =='H') {
								readend+=count;
							}
						}
						if (c[1]!='\0') {
							fprintf(stderr,"Error parsing CIGAR string %s\n",cigar);
							skip=1;
							break;
						}
					}
					count = 0;
				//} else if (*c == 'H'){
				//	//no action necesary
				//	start = FALSE;
				//	count = 0;
				} else if (*c == 'M'){
					start = FALSE;
					//copy from refrence and read
					strncpy(cread,curread,count);
					cread+=count;
					curread+=count;
					strncpy(cgenome,curgenome,count);
					cgenome+=count;
					curgenome+=count;
					count = 0;
				} else if (*c == 'I'){
					start = FALSE;
					//copy from the read
					strncpy(cread,curread,count);
					cread+=count;
					curread+=count;
					memset(cgenome,'-',count);
					cgenome+=count;
					inserts+=count;
					count = 0;
				} else if (*c == 'D'){
					start = FALSE;
					deletions+=count;
					//copy from the refrence
					strncpy(cgenome,curgenome,count);
					cgenome+=count;
					curgenome+=count;
					memset(cread,'-',count);
					cread+=count;
					count=0;
				} else {
					fprintf(stderr,"CIGAR string has unsupported operation %c\n",c);
					skip=1;
					break;	
				}
				c++;
			}
			if (skip==1) {
				continue;	
			}
			rightmost=leftmost+readend-readstart-inserts+deletions;
			*cread='\0';
			*cgenome='\0';
			if (flag & 16 ) {
				reverse(readstring,tempreadstring);
				reverse(genomestring,tempreadstring);
			}
			//*cread='\0';
			//*cgenome='\0';
			//fprintf(stdout,"seq  '%s'\n",seq);
			//fprintf(stdout,"read '%s'\n",readstring);
			//fprintf(stdout,"genome '%s'\n",genomestring);

			//rightmost = leftmost ; readlength - 1;
			write_grp = stderr_grp;


		} else if (line[0] == '>')
		{

			totreads++;

			/*while (*p != 0)
			{
				if (*p == ' ')
					*p = '_';
				p++;
			}*/

			write_grp = stderr_grp;
			char * prev_tab=line;
			char * next_tab=strchr(prev_tab+1,'\t');
			
			sscanf(line, scan_middle,
					readname, contigname, &strand,
					&leftmost, &rightmost, &readstart, &readend,
					&readlength, &score, editstring,
					&normodds, &pgenome, &pchance);
			
			//fprintf(stdout,"DEBUG: readname=%s contigname=%s leftmost=%d rightmost=%d\n",readname, contigname, leftmost, rightmost);
			
			//if (normodds < 0.1)
			//	continue;
		}
		if (useSam == 1 || line[0] == '>')
		{
			if (method == 1) /* this should be the ONLY method ... */
			{
				chr = contigname;
			}
			else if (method == 2)
			{
				int und_pos = -1;
				int j = 0;

				while (contigname[j] != 0)
				{
					if (contigname[j] == '_')
					{
						und_pos = j;
						break;
					}
					j++;
				}

				VA_ASSERT(und_pos >= 0);
				contigname[und_pos] = 0;
				chr2 = atoi(contigname);
				refstart = atoi(&contigname[und_pos + 1]);
			}

			/* determine where this read belongs */
			for (j = 0; j < snpgroup->size; j++)
			{
				if (filenames[j] != NULL)
				{
					SnpGroup_t * s = &snpgroup->data[j];
					int cond = (((leftmost > s->range.data[0]) ||
							(rightmost > s->range.data[0])) &&
							((leftmost < s->range.data[s->range.size - 1]) ||
									(rightmost < s->range.data[s->range.size - 1])));

					if (method == 1)
					{
						cond = cond && (strlen(chr) == s->chr.size);
						cond = cond && (strncmp(chr, s->chr.data, s->chr.size) == 0);
					}
					else if (method == 2)
					{
#if 0
cond = cond && (chr2 == atoi(s->chr)) &&
		(refstart == s->refstart);
#endif
					}

					if (cond)
					{
						if (useSam == 1){
							//fprintf(stdout,"writing lines to write group %i\n",j);
							//fprintf(fids[j],"%s",line);
							fprintf(fids[j],">");
							fprintf(fids[j], scan_middle,
									qname, rname, strand,
									leftmost, rightmost, readstart, readend,
									readlength, 0, "*",
									mapq, mapq, mapq);
							fprintf(fids[j],"\n");

							//write R and G lines.'
							if (color_space==FALSE) {
								uppercase(genomestring);
								uppercase(readstring);
								if (strand=='+') {
									fprintf(fids[j],"G:\t%i\t%s\t%-i\n",leftmost,genomestring,rightmost);
								} else {
									fprintf(fids[j],"G:\t%i\t%s\t%-i\n",rightmost,genomestring,leftmost);
								}
								fprintf(fids[j],"R:\t%i\t%s\t%-i\n",readstart,readstring,readend);
							} else {
								uppercase(genomestring);
								if (strand=='+') {
									fprintf(fids[j],"G:\t%i\t %s\t%-i\n",leftmost,genomestring,rightmost);
								} else {
									fprintf(fids[j],"G:\t%i\t %s\t%-i\n",rightmost,genomestring,leftmost);
								}
								fprintf(fids[j],"T:\t\t %s\n",readstring);
								cs_insert_gaps(color_space_read,readstring,tempreadstring,readstart,readend,readlength);
								fprintf(fids[j],"R:\t%i\t%s\t%-i\n",readstart,color_space_read,readend);
							}

							//write Qline (compose from QUAL)
							if (usequal){
								fprintf(fids[j],"Q:\t");
								char *c = qual;
								while(*c != '\0'){
									fprintf(fids[j],"%i ",*c - 33);
									c++;
								}
								fprintf(fids[j],"\n");
								f_lines[j]+=1;
							}

							write_grp = j;
							if (color_space==FALSE) {
								f_lines[j]+=3;
							} else {
								f_lines[j]+=4;
							}
							f_reads[j]++;
							usedreads++;
							//exit(0);

						} else {
							write_grp = j;
							fprintf(fids[j], "%s", line);
							f_lines[j]++;
							f_reads[j]++;
							usedreads++;
						}
						//fprintf(stdout,"flushing: %i\n\n",fflush(fids[j]));
						break;
					}
				}
			}
		} else
		{
			/* write the line to the previously set file id */
			if (write_grp != stderr_grp)
			{
				fprintf(fids[write_grp], "%s", line);
				f_lines[write_grp]++;
			}
		}
	}

	/* close the file ids. */
	fclose(fid);
	for (i = 0; i < snpgroup->size; i++)
	{
		if (filenames[i] != NULL)
			fclose(fids[i]);
	}

	/* oprint some useful information */
	fprintf(stderr, "\tTotal reads: %d\n", totreads);
	fprintf(stderr, "\tUsed reads: %d\n", usedreads);

	*fid_lines = f_lines;
	*fid_reads = f_reads;
}
